/*#############################################################################
 arrows.c: make PostScript arrows 'figurine' plot -- Marty Sereno
###############################################################################
    1988 -- first hack; rotate x-y coords
90/07/24 -- better parsing
90/08/21 -- penetration points; RF dia icons
90/08/27 -- reads ellipse data, parmfile
90/09/04 -- meridian correction
91/07/21 -- rotate coord system of RFs
91/09/09 -- better legend; real ctx mm (datafile parm); ignore phi>180
92/05/18 -- left/right hemifield
92/08/04 -- write transformed datafile; flip brain hemisphere; landscape output
92/12/10 -- no *.out w/ transformed infile (.out/.str); quits if non-zero trans
93/02/23 -- smaller PostScript output w/defs
93/03/04 -- vector transforms; fix RF-angle (file.out) w/ROTATE_ON_CG
93/05/04 -- setgray bug; upper/lower bug w/flip hemifield
94/01/01 -- fix zero width data (e.g., one arrow)
94/01/23 -- connect-the-penetration-dots => recognizes /break in datafile
94/03/17 -- connect no-response; zero radius merid-correct bug
94/06/20 -- better label number centering
94/11/07 -- also adjust label x centering
14/10/08 -- format, ANSI-ify
17/09/25 -- option to print help
20/08/25 -- embed cmdline/parmfile as comments in ps outfile
#############################################################################*/
/* compile: gcc -D{Linux,Darwin} arrows.c -lm -o arrows */

#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <string.h>

/* default parms: adjust w/parmfile */
#define  NINETY_DEG_ARROW 90.0 /* length (pnts) 90 degree arrow (e.g., 90.0) */
#define  HEAD_SZ      3.5    /* length (points) of arrowhead (e.g., 3.5) */
#define  UPPER        1.2    /* thickness of upper field arrows */
#define  LOWER        0.2    /* thickness of lower field arrows */
#define  FONTSIZE     6.0    /* for point labels */
#define  FRAC_YOFFSET 0.2    /* fraction font yoffset from base of label */
#define  FRAC_XOFFSET 0.0    /* fraction font xoffset from center of label */
#define  PNT_DIA      2.0    /* dia of penetration point */
#define  PNT_LINE     0.2    /* thickness of penetration point line */
#define  PNT_FILL     0.0    /* color of point fill: 0.0-1.0 */
#define  DIA_LINE     0.2    /* thickness of RFdia icon line */
#define  NINETY_DEG_DIA 90.0 /* dia (pnts) 90 deg dia RF icon (e.g., 45.0) */
#define  MERID_CORRECT  0    /* flag turns on VM/HM correction */
#define  MERID_THRESH  0.3   /* promote theta to merid if within this*dia */
#define  MARG         30.0   /* size of margin */
#define  MOVE_ON_HM    0.0   /* rotate around 3-D cartesian y-axis */
#define  MOVE_ON_VM    0.0   /* rotate around 3-D cartesian x-axis */
#define  ROTATE_ON_CG  0.0   /* rotate around 3-D cartesian z-axis */
#define  LEGEND_FLAG     1   /* unset flag to get rid of legend */
#define  LANDSCAPE_FLAG  0   /* set flag get horizontal printout */
#define  LEFT_HEMIFIELD  0   /* set flag to change to left hemifield */
#define  FLIP_HEMISPHERE 0   /* set flag to mirror-image the x-y map */
#define  CONNECT_FLAG    0   /* set flag to connect recording sites */
#define  CONNECT_LINE   0.2  /* thickness of connect line */

/* limits, other defaults */
#define  MAX_ARROWS      20000
#define  HEAD_ANGLE (0.25*M_PI_2) /* angle between stem and one arrow stroke*/
#define  MAX_NAMESIZE     100     /* max num name characters + 1 */
#define  MAX_CORRECT       20.0   /* to catch over-correction w/large RFs */
#define  SCALE_ARROW_DEG   20.0   /* size of arrow pair in legend */
#define  LEGEND_FONT       12.0
#define  TOTAL_PYSIZE     792.0   /* (A4=842.0) page ysize in points */
#define  TOTAL_PXSIZE     612.0   /* (A4=595.0) page xsize in points */
#define  UNITS_PER_MM_CTX  1.0
#define  BLACK    0.0
#define  WHITE    1.0
#define  CONVERT     (M_PI/180.0) /* degrees times this => radians */

/* macros: true for exact match between S and value of variable 'str' */
#define  MATCH_STR(S) (strncmp(str,S,strlen(S))==0)

/* function declarations */
int main(int argc, char **argv);
void ps_header(FILE *fpout);
void ps_parms(FILE *fpout, int argc, char **argv);
void ps_landscape(FILE *fpout);
void ps_arrow(FILE *fpout, float r, float theta, float center_x, float center_y,
              float thickness, float headsize, float headangle);
void ps_text(FILE *fpout, char *str, float x, float y, float fontsize);
void ps_label(FILE *fpout, char *str, float x, float y, float fontsize);
void ps_circle(FILE *fpout, float center_x, float center_y, float dia,
               float thickness, int filled_flag, float filled_color);
void ps_line(
  FILE *fpout, float x1, float y1, float x2, float y2, float thickness);
void ps_rline(
  FILE *fpout, float x, float y, float dx, float dy, float thickness);
void rotate_x(float r[3], float angle);
void rotate_y(float r[3], float angle);
void rotate_z(float r[3], float angle);
void spherpolar_to_cartesian(float r[3]);
void cartesian_to_spherpolar(float r[3]);
float meridian_correct(float radius, float theta, float dia,
                       float merid_thresh, float max_correct);
void print_help(void);


/*** begin main */
int main(int argc, char **argv)
{
  char   line[200], str[200], junk[5];

  int    i, numpts, arrowflag, numflag, diaflag, scaleflag;
  int    legend_flag, landscape_flag, transformed_flag;
  int    merid_correct, left_hemifield, flip_hemisphere;
  int    connect_flag;

  float  ninety_deg_arrow, head_sz, scaled_head, upper, lower, thickness;
  float  marg, pnt_dia, dia_line, ninety_deg_dia, fontsize;
  float  frac_yoffset, frac_xoffset;
  float  parm, rotangle_deg, rotangle, pxsize, pysize;
  float  merid_thresh, units_per_mm_ctx;
  float  x, y, radius_deg, radius, theta_deg, theta, dia_deg, dia;
  float  length_deg, width_deg, angle_deg;
  float  raw_x, raw_y, rot_x, rot_y;
  float  ymax=0.0, xmax=0.0, ymin=0.0, xmin=0.0, xsize, ysize;
  float  x_center_marg, y_center_marg, mm_to_pts, startx, starty;
  float  r[3];
  float  move_on_hm, move_on_vm, rotate_on_cg, tmp;
  float  connect_line;
  float  laststartx=0.0, laststarty=0.0;
  float  pnt_line, pnt_fill;

  FILE   *fopen(), *fpparm, *fpdata, *fpout, *fpndata=NULL;

  typedef struct rf_type_ {
    char   name[MAX_NAMESIZE];
    float  x;
    float  y;
    float  radius;
    float  theta;
    float  dia;
    int    break_before; /* flag */
  } rf_type;
  rf_type  rf[MAX_ARROWS];

  if (argc != 6) {
    printf("\n");
    printf(
    "use: arrows <datafile> <rotate-cw:deg> <arrowflag> <numflag> <diaflag>\n");
    printf(
    "     arrows -h,-help\n");
    printf("\n");
    printf("parms and   NINETY_DEG_ARROW = 90.00   MERID_THRESH = 0.30\n");
    printf("defaults:   HEAD_SZ = 3.50             MARG = 30.00\n");
    printf("            UPPER = 1.20               MOVE_ON_HM = 0.00\n");
    printf("            LOWER = 0.20               MOVE_ON_VM = 0.00\n");
    printf("            FONTSIZE = 6.00            ROTATE_ON_CG = 0.00\n");
    printf("            PNT_DIA = 2.00             LEGEND_FLAG = 1\n");
    printf("            PNT_LINE = 0.2             PNT_FILL = 0.0\n");
    printf("            DIA_LINE = 0.20            LANDSCAPE_FLAG = 0\n");
    printf("            NINETY_DEG_DIA = 90.00     LEFT_HEMIFIELD = 0\n");
    printf("            MERID_CORRECT = 0          FLIP_HEMISPHERE = 0\n");
    printf("            CONNECT_FLAG = 0           CONNECT_LINE = 0.2\n");
    printf("            FRAC_YOFFSET = 0.2         FRAC_XOFFSET = 0.0\n");
    printf("[version 24]\n");
    printf("\n");
    if (argc==2) {
      strcpy(str,argv[1]);
      if ((MATCH_STR("-h")) || (MATCH_STR("-help"))) { print_help(); }
    }
    exit(0);
  }
  rotangle_deg = atof(argv[2]);
  rotangle = -rotangle_deg * CONVERT;  /* neg so points move cw */
  arrowflag = atoi(argv[3]);
  numflag = atoi(argv[4]);
  diaflag = atoi(argv[5]);
  if (arrowflag>1||arrowflag<0||numflag>1||numflag<0||diaflag>1||diaflag<0) {
    printf("arrows: ### bad flag\n");
    exit(1);
  }
  fpdata = fopen(argv[1],"r");
  if (fpdata == NULL)  {
    printf("arrows: ### couldn't find data file: %s\n",argv[1]);
    exit(1);
  }

  transformed_flag = 0;
  i = strlen(argv[1]);
  str[0] = argv[1][i-4];
  str[1] = argv[1][i-3];
  str[2] = argv[1][i-2];
  str[3] = argv[1][i-1];
  str[4] = '\0';
  if (strcmp(str,".out")==0 || strcmp(str,".str")==0)  {
    printf(
    "arrows: ### transformed infile (.out/.str suffix): outfile not written\n");
    transformed_flag = 1;
  }
  if (!transformed_flag)  {
    printf("arrows: transformed data file:  %s.out\n",argv[1]);
    sprintf(str,"%s.out",argv[1]);
    fpndata = fopen(str,"w");
  }

  /* load default parms */
  ninety_deg_arrow = NINETY_DEG_ARROW;
  head_sz = HEAD_SZ;
  upper = UPPER;
  lower = LOWER;
  fontsize = FONTSIZE;
  frac_yoffset = FRAC_YOFFSET;
  frac_xoffset = FRAC_XOFFSET;
  pnt_dia = PNT_DIA;
  pnt_line = PNT_LINE;
  pnt_fill = PNT_FILL;
  dia_line = DIA_LINE;
  ninety_deg_dia = NINETY_DEG_DIA;
  merid_correct = MERID_CORRECT;
  merid_thresh = MERID_THRESH;
  marg = MARG;
  move_on_hm = MOVE_ON_HM;
  move_on_vm = MOVE_ON_VM;
  rotate_on_cg = ROTATE_ON_CG;
  legend_flag = 1;
  landscape_flag = 0;
  left_hemifield = 0;
  flip_hemisphere = 0;
  connect_flag = CONNECT_FLAG;
  connect_line = CONNECT_LINE;
  units_per_mm_ctx = UNITS_PER_MM_CTX;  /* set this one too */

  /* try to open parms file */
  fpparm = fopen("arr.parm","r");
  if (fpparm==NULL) {
    printf("arrows: using default parms; new parmfile is:  arr.parm\n");
    fpparm = fopen("arr.parm","w");
    fprintf(fpparm,"NINETY_DEG_ARROW = %3.2f\n",NINETY_DEG_ARROW);
    fprintf(fpparm,"HEAD_SZ = %3.2f\n",HEAD_SZ);
    fprintf(fpparm,"UPPER = %3.2f\n",UPPER);
    fprintf(fpparm,"LOWER = %3.2f\n",LOWER);
    fprintf(fpparm,"FONTSIZE = %3.2f\n",FONTSIZE);
    fprintf(fpparm,"FRAC_YOFFSET = %3.2f\n",FRAC_YOFFSET);
    fprintf(fpparm,"FRAC_XOFFSET = %3.2f\n",FRAC_XOFFSET);
    fprintf(fpparm,"PNT_DIA = %3.2f\n",PNT_DIA);
    fprintf(fpparm,"PNT_LINE = %3.2f\n",PNT_LINE);
    fprintf(fpparm,"PNT_FILL = %3.2f\n",PNT_FILL);
    fprintf(fpparm,"DIA_LINE = %3.2f\n",DIA_LINE);
    fprintf(fpparm,"NINETY_DEG_DIA = %3.2f\n",NINETY_DEG_DIA);
    fprintf(fpparm,"MERID_CORRECT = %d\n",MERID_CORRECT);
    fprintf(fpparm,"MERID_THRESH = %3.2f\n",MERID_THRESH);
    fprintf(fpparm,"MARG = %3.2f\n",MARG);
    fprintf(fpparm,"MOVE_ON_HM = %3.2f\n",MOVE_ON_HM);
    fprintf(fpparm,"MOVE_ON_VM = %3.2f\n",MOVE_ON_VM);
    fprintf(fpparm,"ROTATE_ON_CG = %3.2f\n",ROTATE_ON_CG);
    fprintf(fpparm,"LEGEND_FLAG = %d\n",LEGEND_FLAG);
    fprintf(fpparm,"LANDSCAPE_FLAG = %d\n",LANDSCAPE_FLAG);
    fprintf(fpparm,"LEFT_HEMIFIELD = %d\n",LEFT_HEMIFIELD);
    fprintf(fpparm,"FLIP_HEMISPHERE = %d\n",FLIP_HEMISPHERE);
    fprintf(fpparm,"CONNECT_FLAG = %d\n",CONNECT_FLAG);
    fprintf(fpparm,"CONNECT_LINE = %3.2f\n",CONNECT_LINE);
    fclose(fpparm);
  }
  else {  /* read any parms in file; others leave default */
    printf("arrows: using **existing** parmfile:\n");
    while (fgets(line,80,fpparm) != NULL) { 
      if (sscanf(line,"%s %s %f",str,junk,&parm) == 3)  {
        if      (MATCH_STR("NINETY_DEG_ARROW")) ninety_deg_arrow = parm;
        else if (MATCH_STR("HEAD_SZ"))       head_sz = parm;
        else if (MATCH_STR("UPPER"))         upper = parm;
        else if (MATCH_STR("LOWER"))         lower = parm;
        else if (MATCH_STR("FONTSIZE"))      fontsize = parm;
        else if (MATCH_STR("FRAC_YOFFSET"))  frac_yoffset = parm;
        else if (MATCH_STR("FRAC_XOFFSET"))  frac_xoffset = parm;
        else if (MATCH_STR("PNT_DIA"))       pnt_dia = parm;
        else if (MATCH_STR("PNT_LINE"))      pnt_line = parm;
        else if (MATCH_STR("PNT_FILL"))      pnt_fill = parm;
        else if (MATCH_STR("DIA_LINE"))      dia_line = parm;
        else if (MATCH_STR("NINETY_DEG_DIA")) ninety_deg_dia = parm;
        else if (MATCH_STR("MERID_CORRECT")) merid_correct = (int)parm;
        else if (MATCH_STR("MERID_THRESH"))  merid_thresh = parm;
        else if (MATCH_STR("MARG"))          marg = parm;
        else if (MATCH_STR("MOVE_ON_HM"))    move_on_hm = parm;
        else if (MATCH_STR("MOVE_ON_VM"))    move_on_vm = parm;
        else if (MATCH_STR("ROTATE_ON_CG"))  rotate_on_cg = parm;
        else if (MATCH_STR("LEGEND_FLAG"))   legend_flag = (int)parm;
        else if (MATCH_STR("LANDSCAPE_FLAG"))  landscape_flag = (int)parm;
        else if (MATCH_STR("LEFT_HEMIFIELD"))  left_hemifield = (int)parm;
        else if (MATCH_STR("FLIP_HEMISPHERE")) flip_hemisphere = (int)parm;
        else if (MATCH_STR("CONNECT_FLAG"))  connect_flag = (int)parm;
        else if (MATCH_STR("CONNECT_LINE"))  connect_line = parm;
        else  printf("arrows: ### don't know this parm: %s  ...ignored\n",str);
      }
      else {
        printf("arrows: ### couldn't parse this line in arr.parm:  %s",line);
        printf("  ...quitting\n");
        exit(1);
      }
    }
    fclose(fpparm);
  }  /* done reading parms */
      
  if (transformed_flag)  {
    if (move_on_hm!=0.0 || move_on_vm!=0.0 || rotate_on_cg!=0.0 ||
        rotangle_deg!=0.0 || merid_correct || flip_hemisphere || 
        left_hemifield)  {
      printf("arrows: ### won't retransform (HM,VM,CG,mer,rot,LH,LF) file:%s\n",
             argv[1]);
      printf("  ...quitting\n");
      exit(1);
    }
  }

  /* get line; skip comments,space,commands; read 6,8args data or 3args parm */
  for (i=0; i<MAX_ARROWS; i++)
    rf[i].break_before = 0;
  i = numpts = 0;
  scaleflag = 0;
  while (fgets(line,80,fpdata) != NULL) { 
    if (line[0]=='#' || line[0]==' ' || line[0]=='\n') /* skip comment,space */
      continue;
    if (line[0]=='/') { /* if command, write flag and continue */
      strcpy(str,line);
      if (MATCH_STR("/break"))
        rf[i].break_before = 1;
      continue;  /* ignore other commands */
    }

    /* if-else block; longest line first! (else later tests skipped) */
    if (sscanf(line,"%s %f %f %f %f %f %f %f",rf[i].name,&raw_x,&raw_y,
               &radius_deg,&theta_deg,
               &length_deg,&width_deg,&angle_deg) == 8) {  /* 8-arg ellipse */
      dia_deg = (length_deg + width_deg)/2.0;
    }
    else if (sscanf(line,"%s %f %f %f %f %f",rf[i].name,&raw_x,&raw_y,
             &radius_deg,&theta_deg,&dia_deg) == 6) {  /* 6-arg oldstyle RF */
      if (dia_deg == 0.0) dia_deg = 5.0;   /* 0dia->std */
      length_deg = width_deg = dia_deg;
    }
    else if (sscanf(line,"%s %s %f",str,junk,&tmp) == 3) {  /* 3-arg parm */
      if (!(MATCH_STR("UNITS_PER_MM_CTX"))) {
        printf("arrows: ### don't know this intrinsic parm in data file:\n");
        printf("   %s",line);
      }
      else  {
        units_per_mm_ctx = tmp;
        scaleflag = 1;
      }
      continue;  /* don't continue processing this line */
    }
    else  {  /* err */
      printf("arrows: ### couldn't parse this line in data file:\n");
      printf("   %s",line);
      continue;
    }

    if (raw_x==9999.0 && raw_y==9999.0) continue;  /* for no-arrow B.S. */

    if (flip_hemisphere) raw_x = -raw_x;  /* before rot, max/min */

    rot_x = raw_x*cos((double)rotangle) - raw_y*sin((double)rotangle); /* rot */
    rot_y = raw_x*sin((double)rotangle) + raw_y*cos((double)rotangle);

    x = rot_x/units_per_mm_ctx;   /* to cortical mm */
    y = rot_y/units_per_mm_ctx;

    if (i==0) { ymax=ymin=y; xmax=xmin=x; }  /* max/min */
    if (x>xmax) xmax=x; if (y>ymax) ymax=y;
    if (x<xmin) xmin=x; if (y<ymin) ymin=y;
    rf[i].x = x;
    rf[i].y = y;

    if (radius_deg > 180.0) {  /* ignore rad (phi,latitude) over 180 */
      printf("arrows: ### radius over 180 deg--line ignored\n");
      printf("   %s",line);
      continue;
    }

    radius = radius_deg*CONVERT;  /* to radians */
    theta = theta_deg*CONVERT;
    dia = dia_deg*CONVERT;

    /* transform RF coords to cartesian; rotate; transform back */
    /* NOTE: my "radius"= spher. polar phi, "theta"= spher. polar theta */
    /* NOTE: rotations are like: x-along-HM, y-along-VM, rotate-around-CG */
    /* signs: rot,vm -> no neg (cw = neg angle, pos x is down [to viewer] */
    if (radius>0.0) {  /* don't transform NR's */
      r[0] = 1.0;
      r[1] = theta;
      r[2] = radius;
      spherpolar_to_cartesian(r);
      rotate_y(r,-move_on_hm*CONVERT);
      rotate_x(r,move_on_vm*CONVERT);
      rotate_z(r,rotate_on_cg*CONVERT);
      cartesian_to_spherpolar(r);
      theta = r[1];
      radius = r[2];
    }

    /* retain neg theta for angles in -PI/2 to -PI (so arrows stay lower) */
    if (theta > M_PI && theta < 3.0*M_PI_2+0.000001)
      theta -= 2.0*M_PI;

    /* RFdia and eccen. depend. promotion of theta near VM and HM toward them */
    if (merid_correct && radius>0.0)
      theta = meridian_correct(radius,theta,dia,merid_thresh,MAX_CORRECT);
	
    /* flip hemisphere; fix RF-angle */
    if (left_hemifield) {  /* NOTE: coord trans only rotates center */ 
      theta = M_PI - theta;
      angle_deg += rotate_on_cg;
    }
    else
      angle_deg -= rotate_on_cg;

    /* save in arrays */
    rf[i].radius = radius;
    rf[i].theta = theta;
    rf[i].dia = dia;

    /* write transformed outfile line (x-y in rotated mm; r,theta transformed)*/
    if (!transformed_flag) {
      fprintf(fpndata,"%s\t%3.4f\t%3.4f\t%3.3f\t%3.3f\t%3.1f\t%3.1f\t%3.1f\n",
              rf[i].name,x,y,radius/CONVERT,theta/CONVERT,
              length_deg,width_deg,angle_deg);
    }
    i++;
    if (i>=MAX_ARROWS){
      printf("arrows: ### overflowed %d points)\n",MAX_ARROWS);
      exit(1); 
    }
  }  /* data file lines */
  fclose(fpdata);
  if (!transformed_flag) fclose(fpndata);

  /* finish up */
  if (scaleflag)
    printf("arrows: %3.2f x-y units = 1.0 mm cortex\n",units_per_mm_ctx);
  else {
    printf("arrows: no cortex scale parm (UNITS_PER_MM_CTX) in datafile:\n");
    printf("  ...assuming %3.1f unit = 1.0 mm ctx\n",units_per_mm_ctx);
  }
  numpts = i;
  ysize = ymax - ymin;
  xsize = xmax - xmin;
  if (ysize==0.0)  { ysize=1.0; ymax=0.5; ymin= -0.5; }
  if (xsize==0.0)  { xsize=1.0; xmax=0.5; xmin= -0.5; }
  /*** done with data file ***/

  /* working coords */
  if (landscape_flag) {  /* flip x and y */
    pysize = TOTAL_PXSIZE - (2.0*marg);
    pxsize = TOTAL_PYSIZE - (2.0*marg);
  }
  else {
    pxsize = TOTAL_PXSIZE - (2.0*marg);
    pysize = TOTAL_PYSIZE - (2.0*marg);
  }

  /* see if x or y goes to edge; get scale factor, x-y center-margins */
  if (ysize/xsize > pysize/pxsize) {
    mm_to_pts = pysize/ysize; /* mult cortex x-y by this to get points */
    x_center_marg = (pxsize - xsize*mm_to_pts)/2.0;
    y_center_marg = 0.0;  }
  else {
    mm_to_pts = pxsize/xsize;
    x_center_marg = 0.0;
    y_center_marg = (pysize - ysize*mm_to_pts)/2.0;
  }

  /* open outfile and write PostScript header */
  sprintf(str,"%s.arr.ps",argv[1]);
  printf("arrows: PostScript outfile:  %s\n",str);
  fpout = fopen(str,"w");
  ps_header(fpout);
  ps_parms(fpout,argc,argv);
  if (landscape_flag)
    ps_landscape(fpout);

  if (legend_flag) {

    /* write filename, parms at upper left */
    sprintf(str,"file:  %s",argv[1]);
    ps_text(fpout,str,marg,pysize+marg-0.0*LEGEND_FONT,LEGEND_FONT);
    sprintf(str,"map rotated  %d degrees",(int)rotangle_deg);
    ps_text(fpout,str,marg,pysize+marg-1.0*LEGEND_FONT,LEGEND_FONT);
    if (arrowflag) {
      if (merid_correct)
        ps_text(fpout,"meridian correction",
                marg,pysize+marg-2.0*LEGEND_FONT,LEGEND_FONT);
      else
        ps_text(fpout,"no meridian correction",
                marg,pysize+marg-2.0*LEGEND_FONT,LEGEND_FONT);
    }

    /* write arrow legend at upper left below parms */
    radius = ninety_deg_arrow*((SCALE_ARROW_DEG*CONVERT)/M_PI_2);
    startx = marg + 8.75*LEGEND_FONT;  /* after text */
    starty = marg + pysize - 2.5*LEGEND_FONT - radius/2.0; /* arrow center */
    if (left_hemifield)  /* move VM arrow over */
      ps_arrow(
       fpout,radius,90.0*CONVERT,startx+radius,starty,lower,head_sz,HEAD_ANGLE);
    else
      ps_arrow(
        fpout,radius,90.0*CONVERT,startx,starty,lower,head_sz,HEAD_ANGLE);
    sprintf(str,"%d degree arrows:",(int)SCALE_ARROW_DEG);
    ps_text(fpout,str,marg,starty-LEGEND_FONT/2.0,LEGEND_FONT);/*centertext*/
    startx += radius/2.0;  /* HM arrow */
    starty -= radius/2.0;
    if (left_hemifield) theta = M_PI - 0.0*CONVERT;
    else                theta = 0.0*CONVERT;
    ps_arrow(fpout,radius,theta,startx,starty,lower,head_sz,HEAD_ANGLE);

    /* write cortex-coords scalebar w/centered (1.33) text below arrow legend*/
    startx = marg + LEGEND_FONT + (1.0*mm_to_pts)/2.0 - 1.33*LEGEND_FONT;
    starty = marg + pysize - 2.5*LEGEND_FONT - radius/2.0 - 2.0*LEGEND_FONT;
    ps_text(fpout,"1 mm",startx,starty,LEGEND_FONT);
    startx = marg + LEGEND_FONT;  /* start one char in */
    starty -= 3.0;                /* put line close below */
    ps_rline(fpout,startx,starty,1.0*mm_to_pts,0.0,2.0);  /* scale bar */

    /* write transformation legend at upper right */
    if (move_on_hm!=0.0 || move_on_vm!=0.0 || rotate_on_cg!=0.0)  {
      sprintf(str,"RF transformations");
      ps_text(
        fpout,str,marg+0.75*pxsize,marg+pysize-0.0*LEGEND_FONT,LEGEND_FONT);
      sprintf(str,"HM:  %d deg",(int)move_on_hm);
      ps_text(
        fpout,str,marg+0.77*pxsize,marg+pysize-1.0*LEGEND_FONT,LEGEND_FONT);
      sprintf(str,"VM:  %d deg",(int)move_on_vm);
      ps_text(
        fpout,str,marg+0.77*pxsize,marg+pysize-2.0*LEGEND_FONT,LEGEND_FONT);
      sprintf(str,"Rot:  %d deg",(int)rotate_on_cg);
      ps_text(
        fpout,str,marg+0.77*pxsize,marg+pysize-3.0*LEGEND_FONT,LEGEND_FONT);
    }
  }  /* if legend_flag */

  /*** drawing loop each point */
  for (i=0; i<numpts; i++)  {

    /* find center point for arrow (pagemargin,center-margs,x,ymins) */
    starty = marg + y_center_marg + (rf[i].y - ymin)*mm_to_pts;
    startx = marg + x_center_marg + (rf[i].x - xmin)*mm_to_pts;
    if (i==0) { laststartx=startx; laststarty=starty; }

    /* draw arrow, no-resp point, else bare point */
    if (arrowflag) {  
      if (rf[i].radius>0.0) {  /* if response, draw arrow */
        radius = ninety_deg_arrow*(rf[i].radius/M_PI_2);
        if (rf[i].theta>=0.0 && rf[i].theta<=M_PI) thickness = upper;
        else                                       thickness = lower;
        if (radius<2.0*head_sz)  {  /* scale head on small arrows */
          scaled_head = 0.6*radius;
          if (scaled_head<1.0) scaled_head = 1.0;  /* real-small cutoff */
        }
        else scaled_head = head_sz;
        ps_arrow(fpout,radius,rf[i].theta,startx,starty,thickness,
                                                       scaled_head,HEAD_ANGLE);
      }  /* if response */
      else  /* draw no-resp point */
        ps_circle(fpout,startx,starty,pnt_dia,pnt_line,1,BLACK);
    }
    else
      ps_circle(fpout,startx,starty,pnt_dia,pnt_line,1,BLACK); /* bare point */

    /* draw numbers */
    if (numflag && arrowflag)
      ps_text(fpout,rf[i].name,startx+pnt_dia/2.0,starty-fontsize/2.0,fontsize);
    if (numflag && !arrowflag)
      ps_label(fpout,rf[i].name,startx + frac_xoffset*fontsize,
               starty + frac_yoffset*fontsize + pnt_dia/2.0,fontsize);

    /* draw unfilled RFdia icons (only if arrow) */
    if (diaflag && rf[i].radius>0.0) {
      dia = ninety_deg_dia*(rf[i].dia/M_PI_2);  /* dia at 90 degrees */
      ps_circle(fpout,startx,starty,dia,dia_line,0,BLACK);
    }

    /* draw lines if needed (occlude line by redrawing current and last point */
    if (connect_flag) {
      if (!rf[i].break_before && i) {
        ps_line(fpout,startx,starty,laststartx,laststarty,connect_line);
        ps_circle(fpout,startx,starty,pnt_dia,pnt_line,1,pnt_fill);
        ps_circle(fpout,laststartx,laststarty,pnt_dia,pnt_line,1,pnt_fill);
      }
      laststartx = startx;
      laststarty = starty;
    }
  }  /* arrows/points */

  /* Postscript trailer */
  fprintf(fpout,"showpage\n");
  fclose(fpout);
  exit(0);
}

/*** begin functions */
void ps_header(FILE *fpout)
{
  fprintf(fpout,"%%!PS-Adobe-2.0\n");
  fprintf(fpout,"%%%%BoundingBox: 0.0 0.0 612.0 792.0\n");
  fprintf(fpout,"1.00 1.00 scale\n");
  fprintf(fpout,"newpath\n");
  fprintf(fpout,"0.00 0.00 translate\n");
  fprintf(fpout,"0.00 setgray\n");    /* start off drawing in black */

  /* defs for a smaller output file */
  fprintf(fpout,"/a { gsave } def\n");      /* added 1/23/94 */
  fprintf(fpout,"/e { grestore } def\n");   /* added 1/23/94 */
  fprintf(fpout,"/f { fill } def\n");
  fprintf(fpout,"/g { setgray } def\n");    /* added 1/23/94 */
  fprintf(fpout,"/l { lineto } def\n");     /* added 1/23/94 */
  fprintf(fpout,"/L { rlineto } def\n");
  fprintf(fpout,"/m { moveto } def\n");
  fprintf(fpout,"/n { newpath } def\n");
  fprintf(fpout,"/r { rotate } def\n");
  fprintf(fpout,"/s { stroke } def\n");
  fprintf(fpout,"/S { show } def\n");
  fprintf(fpout,"/t { translate } def\n");
  fprintf(fpout,"/w { setlinewidth } def\n");
  fprintf(fpout,"/fh { /Helvetica findfont } def\n");
  fprintf(fpout,"/ss { scalefont setfont } def\n");
  fprintf(fpout,"/sl { stringwidth pop 0 exch sub } def\n"); /* 6/20/94 */
}

void ps_parms(FILE *fpout, int argc, char **argv)
{
  int   i;
  char  line[200];
  FILE  *fpparm=NULL;

  fprintf(fpout,"%% cmdline:");
  for (i=0; i<argc; i++)
    fprintf(fpout," %s",argv[i]);
  fprintf(fpout,"\n");

  fpparm = fopen("arr.parm","r");
  if (fpparm==NULL) {
    fprintf(fpout,"%% no arr.parm: used all-default parms\n");
  }
  else {
    fprintf(fpout,"%% arr.parm (non-default parms):\n");
    while (fgets(line,80,fpparm) != NULL) {
      if (strlen(line) && line[0]!='\n') {
        fprintf(fpout,"%% %s",line);     /* fgets keeps newline */
        if (line[strlen(line)-1]!='\n')  /* if last missing */
          fprintf(fpout,"\n");       
      }
    }
  }
}

void ps_landscape(FILE *fpout)
{
  fprintf(fpout,"90.00 r\n");
  fprintf(fpout,"0.00 -612.00 t\n");
}

void ps_arrow(FILE *fpout, float r, float theta, float center_x, float center_y,
              float thickness, float headsize, float headangle)
{
  float  start_stem_x, start_stem_y;
  float  stem_dx, stem_dy, halfstem_dx, halfstem_dy;
  float  r_head_dx, r_head_dy, l_head_dx, l_head_dy;
  float  r_head_x, r_head_y;

  /* dx,dy for arrowstem; half arrow since we rotate around */
  halfstem_dx = (r/2.0)*(cos((double)theta));
  halfstem_dy = (r/2.0)*(sin((double)theta));

  /* make arrow rotate around center_x,center_y */
  start_stem_x = center_x - halfstem_dx;
  start_stem_y = center_y - halfstem_dy;
  stem_dx = 2.0*halfstem_dx;
  stem_dy = 2.0*halfstem_dy;

  /* arrowhead (+30,-30 degrees => +150,+210 rel to arrow tip) */
  r_head_dx = headsize*cos((double)(theta + M_PI - headangle));
  r_head_dy = headsize*sin((double)(theta + M_PI - headangle));
  l_head_dx = headsize*cos((double)(theta + M_PI + headangle));
  l_head_dy = headsize*sin((double)(theta + M_PI + headangle));

  /* abs coord right arrowhead end */
  r_head_x = start_stem_x + stem_dx + r_head_dx;
  r_head_y = start_stem_y + stem_dy + r_head_dy;

  /* write PostScript */
  fprintf(fpout,"n\n");
  fprintf(fpout,"%1.2f w\n",thickness);

  /* draw tail -> tip */
  fprintf(fpout,"%3.2f %3.2f m\n",start_stem_x,start_stem_y);
  fprintf(fpout,"%3.2f %3.2f L\n",stem_dx,stem_dy);

  /* draw rightend -> tip -> leftend (makes sharp point) */
  fprintf(fpout,"%3.2f %3.2f m\n",r_head_x,r_head_y);
  fprintf(fpout,"%3.2f %3.2f L\n",-r_head_dx,-r_head_dy);
  fprintf(fpout,"%3.2f %3.2f L\n",l_head_dx,l_head_dy);
  fprintf(fpout,"s\n");
}

void ps_text(FILE *fpout, char *str, float x, float y, float fontsize)
{
  fprintf(fpout,"fh\n");
  fprintf(fpout,"%d ss\n",(int)fontsize);
  fprintf(fpout,"%3.2f %3.2f m\n",x,y);
  fprintf(fpout,"(%s) S\n",str);
}

void ps_label(FILE *fpout, char *str, float x, float y, float fontsize)
{
  fprintf(fpout,"fh\n");
  fprintf(fpout,"%d ss\n",(int)fontsize);
  fprintf(fpout,"(%s) sl\n",str);         /* x (-stringlength) on stack */
  fprintf(fpout,"%3.2f mul\n",0.5);       /* center -x on stack */
  fprintf(fpout,"%3.2f add\n",x);         /* absolute x on stack */
  fprintf(fpout,"%3.2f m\n",y);           /* push y, go there */
  fprintf(fpout,"(%s) S\n",str);
}

void ps_circle(FILE *fpout, float center_x, float center_y, float dia,
               float thickness, int filled_flag, float filled_color)
{
  if (dia==0.0) return;
  fprintf(fpout,"n\n");                   /* PS: newpath */
  fprintf(fpout,"%1.2f w\n",thickness);   /* PS: setlinewidth */
  fprintf(fpout,"%3.2f %3.2f %3.2f 0.0 360.0 arc\n",center_x,center_y,dia/2.0);
  if (filled_flag)  {
    fprintf(fpout,"%3.2f g\n",filled_color);  /* PS: setgray */
    fprintf(fpout,"a\n");       /* PS: gsave (save path for stroke) */
    fprintf(fpout,"f\n");       /* PS: fill */
    fprintf(fpout,"e\n");       /* PS: grestore */
    fprintf(fpout,"0.0 g\n");   /* PS: setgray (undo for next) */
  } 
  fprintf(fpout,"s\n");         /* PS: stroke */
}

void ps_line(
  FILE *fpout, float x1, float y1, float x2, float y2, float thickness)
{
  fprintf(fpout,"n\n");
  fprintf(fpout,"%1.2f w\n",thickness);
  fprintf(fpout,"%3.2f %3.2f m\n",x1,y1);
  fprintf(fpout,"%3.2f %3.2f l\n",x2,y2);
  fprintf(fpout,"s\n");
}

void ps_rline(
  FILE *fpout, float x, float y, float dx, float dy, float thickness)
{
  fprintf(fpout,"n\n");
  fprintf(fpout,"%1.2f w\n",thickness);
  fprintf(fpout,"%3.2f %3.2f m\n",x,y);
  fprintf(fpout,"%3.2f %3.2f L\n",dx,dy);
  fprintf(fpout,"s\n");
}

void rotate_x(float r[3], float angle)  /* rotate vect around x-axis radians */
{
  float  y, z;

  y = r[1];
  z = r[2];
  r[1] = y*cos((double)angle) + z*sin((double)angle);
  r[2] = -y*sin((double)angle) + z*cos((double)angle);
}

void rotate_y(float r[3],float angle)  /* rotate vect around y-axis radians */
{
  float  x, z;

  x = r[0];
  z = r[2];
  r[0] = x*cos((double)angle) - z*sin((double)angle);
  r[2] = x*sin((double)angle) + z*cos((double)angle);
}

void rotate_z(float r[3], float angle)  /* rotate vect around z-axis radians */
{
  float  x, y;

  x = r[0];
  y = r[1];
  r[0] = x*cos((double)angle) + y*sin((double)angle);
  r[1] = -x*sin((double)angle) + y*cos((double)angle);
}

void spherpolar_to_cartesian(float r[3])  /* spherpol (rho,th,phi) -> x,y,z */
{
  float  rho, theta, phi;

  rho = r[0];
  theta = r[1];
  phi = r[2];   /* remember 0 < phi < PI */
  r[0] = (double)rho * sin((double)phi) * cos((double)theta);
  r[1] = (double)rho * sin((double)phi) * sin((double)theta);
  r[2] = (double)rho * cos((double)phi);
}

void cartesian_to_spherpolar(float r[3])  /* x,y,z -> spherpol (rho,th,phi) */
{
  float  x, y, z, rho, theta=0.0, phi;

  x = r[0];
  y = r[1];
  z = r[2];
  rho = sqrt((double)(x*x + y*y + z*z));
  if (rho==0.0)
    theta = phi = 0.0;
  else {
    if (x>0.0)
      theta = atan((double)(y/x));
    if (x<0.0)
      theta = M_PI + atan((double)(y/x));
    if (x==0.0 && y>=0.0)
      theta = M_PI_2;
    if (x==0.0 && y<0.0)
      theta = 3.0 * M_PI_2;
    phi = acos((double)(z/(rho)));
    }
  r[0] = rho;
  r[1] = theta;
  r[2] = phi;
}

/*============================================================================
 Promote theta near VM,HM toward them as function of eccen (rad) and RF dia
 (see parms, in degrees).  Expects neg theta's for Pi to 3Pi/2 and radius>0.
============================================================================*/
float meridian_correct(float radius, float theta, float dia,
                       float merid_thresh, float max_correct)
{
  float  theta_deg, window_sine, max_corr_deg, corr_sector_deg;

  if (radius==0.0) return theta;
  theta_deg = theta/CONVERT;  /* un-convert theta for correction */
  window_sine = (dia*merid_thresh)/radius;
  if (window_sine > 1.0)           /* catch lg dia with small r */
    window_sine = 0.99;
  max_corr_deg = (asin((double)window_sine))/CONVERT;
  if (max_corr_deg > max_correct)  /* catch over-correction from lg RFs */
    max_corr_deg = max_correct; 
  corr_sector_deg = 45.0 - max_corr_deg; /* distrib corr over this sector */
  if (theta_deg <= 135.0 && theta_deg > 90.0) {        /* to +VM */
    theta_deg -= max_corr_deg*((135.0-theta_deg)/corr_sector_deg);
    if (theta_deg < 90.0)   theta_deg = 90.0;
  }
  else if (theta_deg <= 90.0 && theta_deg > 45.0) {    /* to +VM */
    theta_deg += max_corr_deg*((theta_deg-45.0)/corr_sector_deg);
    if (theta_deg > 90.0)   theta_deg = 90.0;
  }
  else if (theta_deg <= 45.0 && theta_deg > 0.0) {     /* to HM */
    theta_deg -= max_corr_deg*((45.0-theta_deg)/corr_sector_deg);
    if (theta_deg < 0.0)   theta_deg = 0.0001;
  }
  else if (theta_deg <= 0.0 && theta_deg > -45.0) {    /* to HM */
    theta_deg += max_corr_deg*((theta_deg+45.0)/corr_sector_deg);
    if (theta_deg > 0.0)   theta_deg = 0.0;
  }
  else if (theta_deg <= -45.0 && theta_deg > -90.0) {  /* to -VM */
    theta_deg -= max_corr_deg*((-45.0-theta_deg)/corr_sector_deg);
    if (theta_deg < -90.0)   theta_deg = -90.0;
  }
  else if (theta_deg <= -90.0 && theta_deg > -135.0) { /* to -VM */
    theta_deg += max_corr_deg*((theta_deg+135.0)/corr_sector_deg);
    if (theta_deg > -90.0)   theta_deg = -90.0;
  }
  else {
   printf("arrows: RF w/theta=%3.2f (beyond -135,135) uncorrected\n",theta_deg);
  }
  return(theta_deg*CONVERT);  /* re-convert */
}

void print_help(void)
{
printf("\
#############################################################################\n\
              arrows -- make PostScript arrows 'figurine' plot\n\
\n\
                       written by Marty Sereno\n\
                  University of California, San Diego\n\
#############################################################################\n\
    Reads ASCII data file of penetration coordinates and RF params\n\
  and writes Postscript file with drawing of recording sites.  Main use\n\
  is to draw arrow icons indicating the receptive field position.  First\n\
  parameter allows rotation of the coordinate system of the cortex (to\n\
  allow the penetration rows to be aligned with the vertical or horizontal\n\
  axes so the arrow angle is easier to parse).  A positive argument rotates\n\
  the map clockwise.  First flag gives arrows icons at each recording site\n\
  vs. just points; second flag tells whether to number the arrows or\n\
  recording sites; third flag indicates whether to draw circle icons to\n\
  indicate relative RF diameter.\n\
\n\
    Format for data line in input file (which is also read by\n\
  rfplots.c) is either 6 or 8 args:\n\
\n\
pointname  point-x  point-y  RF-r  RF-theta  RF-dia       \n\
                        *or*\n\
pointname  point-x  point-y  RF-r  RF-theta  RF-length  RF-width  RF-angle\n\
\n\
    RF-r (distance from center of gaze to receptive field center)\n\
  given in 0-90+ degrees and RF-theta (angle of center-of-gaze-to-RF\n\
  line) in -90 to +90 degrees, RF-dia (or RF-length, RF-width)\n\
  given in positive degrees and RF-angle (measured between the long \n\
  axis of the receptive field and horizontal) given in -90 to +90 degrees.\n\
  There can be up to 10 chars in point name.\n\
\n\
    Right and left hemifield data should be entered with the horizontal\n\
  meridian as zero degrees.  Default is to draw arrows (and receptive)\n\
  fields as a right hemifield.  If left hemifield is desired, set\n\
  LEFT_HEMIFIELD flag.  If mirror-image of the recording sites is desired\n\
  (for comparing opposite hemispheres), set FLIP_HEMISPHERE flag (individual\n\
  arrows are not flipped--only translated).\n\
\n\
    Lines with leading pound sign or space are treated as comments\n\
  and skipped (i.e., data or command lines *must* have a character \n\
  in the first column).  Can't deal with comments beginning midline.\n\
\n\
    Commands have leading slash.  /black, /dark, /light are currently\n\
  ignored.  /break here means a break in the connection between the\n\
  recording sites (when CONNECT_FLAG is set).  Note that /break here\n\
  breaks recording site connections while /break in rfplots breaks\n\
  RF connections.\n\
#\n\
# Example data file\n\
# \n\
/black\n\
12F	5.23	9.25	24	-25	7.5	4.5	50\n\
254-L	2.31	5.36	37	50	5.5	3.5	-60\n\
312G	10.55	4.46	10	-10	7.5	5	-40\n\
#\n\
# no-response point (no RF; point but no arrow)\n\
25NR	6.24	10.21	-1	0	0	0	0\n\
#\n\
#\n\
/break\n\
/light\n\
571	3.24	7.35	12	57	14	9	-27\n\
601B	6.01	10.05	17	-9	10	8	-5\n\
#\n\
# blind spot 'receptive field' (RF; no point or arrow)\n\
BS-1	9999.0	9999.0	21	-5	4	2	90\n\
#\n\
#\n\
    Also reads parm file (arr.parm) file containing various settings\n\
  (e.g., margins, postscript lines sizes, coordinate transforms,\n\
  meridian correct flag); if not there writes a default version of\n\
  the file, which can then be edited.  Ignores properly formatted \n\
  parameters in parm file that it doesn't recognize.  If a parm appears\n\
  more than once, the last instance determines its value.  Parms omitted\n\
  from file are set to default values.  Cmdline and non-default parms are\n\
  written as comments at head of PostScript outfile. Default parm file:\n\
\n\
NINETY_DEG_ARROW = 90.00\n\
HEAD_SZ = 3.50\n\
UPPER = 1.20\n\
LOWER = 0.20\n\
FONTSIZE = 6.00\n\
FRAC_YOFFSET = 0.2\n\
FRAC_XOFFSET = 0.2\n\
PNT_DIA = 2.00\n\
PNT_LINE = 0.2\n\
PNT_FILL = 0.0\n\
DIA_LINE = 0.20\n\
NINETY_DEG_DIA = 90.00\n\
MERID_CORRECT = 0\n\
MERID_THRESH = 0.30\n\
MARG = 30.00\n\
MOVE_ON_HM = 0.00\n\
MOVE_ON_VM = 0.00\n\
ROTATE_ON_CG = 0.00\n\
LEGEND_FLAG = 1\n\
LANDSCAPE_FLAG = 0\n\
LEFT_HEMIFIELD = 0\n\
FLIP_HEMISPHERE = 0\n\
CONNECT_FLAG = 0\n\
CONNECT_LINE = 0.2\n\
\n\
    Calculates arrows, autoscales, centers, generates PS-Adobe-2.0 format\n\
  Postscript file.  Accepts up to 20000 arrows.  Can rotate the coord\n\
  system from command line so that penetration rows can be aligned with\n\
  the horizontal, making it easier to see the absolute values of arrow\n\
  params (arrow angles are drawn relative to page edges).  Writes out\n\
  the penetration number in FONTSIZE pt Helvetica if requested.  MARG\n\
  sets margin; legend is inside margin.  Set CONNECT_FLAG to connect\n\
  penetration points with CONNECT_LINE thickness (recognizes /break\n\
  command).  PNT_DIA penetration points drawn as PNT_LINE thick circles\n\
  with PNT_FILL color fill and occlude connnect lines.  FRAC_YOFFSET\n\
  adjust base and FRAC_XOFFSET adjusts center of label for point plots\n\
  (no effect on arrow labels).\n\
\n\
    Put in a neg number for r (and fill remaining entries with 0's)\n\
  to get a point with a number but no RF arrow (e.g., injection,\n\
  non-recorded lesion, no response).  Put in 9999.0 for x and y to\n\
  suppress arrow (for RFs for blind spot measurments).\n\
\n\
    MERID_CORRECT flag promotes all RF centers within a certain\n\
  distance (MERID_THRESH * RF-dia) of the vertical and horizontal\n\
  meridians to them; meridian promotion falls to zero at -45 and\n\
  45 degrees.  MAX_CORRECT is largest allowed correction (to catch\n\
  over-correction with large RFs).\n\
\n\
    Can rotate coord system of RFs.  This is done in the following order:\n\
  move along the HM (rotate around cartesian y-axis), move along the\n\
  VM (rotate around cartesian x-axis), and rotate around the center of\n\
  gaze (rotate around cartesian z-axis).  These transformations are\n\
  done *before* optional hemifield flip.\n\
  \n\
    Writes transformed datafile (datafile.out) omitting comments but\n\
  containing no-response points (neg radius) so that these can be\n\
  stretched as well as RF points.  Cortex x-y is written in rotated mm\n\
  and r-theta are rotated, hemisphere and hemifield flipped, and \n\
  meridian-corrected.  Doesn't write transformed outfile if datafile\n\
  appears to already be transformed (.out or .str suffix).  Quits\n\
  if non-zero transformations requested (rotangle, or in arr.parm) for\n\
  transformed input datafile.\n\
\n\
    Looks for 3 arg parmfile-like line in **datafile**, for example:\n\
\n\
UNITS_PER_MM_CTX = 2.0\n\
\n\
  indicating ratio between datafile x-y units and real cortex mm.  This\n\
  parm must come before data points to which it applies.  Default = 1.0.\n\
  Unset LEGEND_FLAG to get rid of legend and set LANDSCAPE_FLAG to get\n\
  horizontal output.\n\
#############################################################################\n\
\n\
");
}

