/*#############################################################################
 fieldsign.c: calc visual fieldsign (mirror vs. non-mirror) -- Marty Sereno
    1992 -- first hack
92/01/13 -- working
92/01/15 -- coloring options
93/03/16 -- output x,y,z file for GMT
93/05/13 -- parmfile
94/05/22 -- fixed ignoring parmfile
94/06/04 -- added hemisphere parm (since theta is always  -90 => 90!!)
94/06/11 -- put in sigmoid color transitions
14/10/08 -- ANSI-ify
17/09/25 -- option to print help
#############################################################################*/
/* compile: gcc -D{Linux,Darwin} fieldsign.c -lm -o fieldsign */

#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <string.h>

/* color table */
#define  MIN_DIFF  (0.0 * M_PI)  /* region of BETWEEN; e.g., (0.2*M_PI) */
#define  NON_MIRROR       100    /* non-mirror-image (100=dark gray) */
#define  BETWEEN          175    /* can't tell (white) */
#define  MIRROR           200    /* mirror-image color (200=light gray) */
#define  BORDER           250    /* white */
#define  LEFT_HEMISPHERE    0    /* set flag if data from left hemisphere */
#define  FLIPPED_HEMISPHERE 0    /* set flag if arrows flipped hemisphere */
#define  SIGMOID_SLOPE     4.0   /* adjust width of transition zone */

#define  CONVERT   (2.0*M_PI/360.0)      /* degrees times this ==> radians */

/* true for exact match between S and str */
#define  MATCH_STR(S)   (strncmp(str,S,strlen(S)) == 0)

/* color options */
#define  GRAYSCALE   0
#define  BINARY      1
#define  SIGMOID     2

int main(int argc, char **argv);
void print_help(void);


/*** begin main */
int main(int argc, char **argv)
{
  char   str[81], line[81], junk[5];

  int    i, x, y, xi, yi, xsize, ysize;
  int    non_mirror, between, mirror, border;
  int    tmp, left_hemisphere, flipped_hemisphere;
  int    color=0, color_option, offset;

  float  sigmoid_slope, sig;
  float  sense_angle, radius_grad_angle, theta_grad_angle, min_diff;
  float  radius_deg, theta_deg, dx_radius, dy_radius, dx_theta, dy_theta;
  float  **radius_array, **theta_array;
  float  parm;

  FILE   *fopen(), *fpsign, *fpradius, *fptheta, *fpparm;

  if (argc != 6)  {
    printf("\n");
    printf(
"use: fieldsign <rootfile> <xsize> <ysize> <offset> <colopt=0gry,1bin,2sig>\n");
    printf("\n");
    printf("parms and      MIN_DIFF = %3.2f         MIRROR = %d\n",
      MIN_DIFF,MIRROR);
    printf("defaults:      NON_MIRROR = %d        BORDER = %d\n",
      NON_MIRROR,BORDER);
    printf("               BETWEEN = %d           LEFT_HEMISPHERE = %d\n",
      BETWEEN,LEFT_HEMISPHERE);
    printf("               FLIPPED_HEMISPHERE = %d  SIGMOID_SLOPE = %3.2f \n",
      FLIPPED_HEMISPHERE,SIGMOID_SLOPE);
    printf("[version 10]\n");
    printf("\n");
    if (argc==2) {
      strcpy(str,argv[1]);
      if ((MATCH_STR("-h")) || (MATCH_STR("-help"))) { print_help(); }
    }
    exit(0);
  }
  xsize = atoi(argv[2]);
  ysize = atoi(argv[3]);
  offset = atoi(argv[4]);
  color_option = atoi(argv[5]);
  if (color_option<0 || color_option>2) {
    fprintf(stderr,"fieldsign: ### bad color option   ...quitting\n");
    exit(1);
  }

  /* load default parms */
  min_diff = MIN_DIFF;
  non_mirror = NON_MIRROR;
  between = BETWEEN;
  mirror = MIRROR;
  border = BORDER;
  left_hemisphere = LEFT_HEMISPHERE;
  flipped_hemisphere = FLIPPED_HEMISPHERE;
  sigmoid_slope = SIGMOID_SLOPE;

  /* try to open parms file */
  fpparm = fopen("field.parm","r");
  if (fpparm==NULL) {
    printf("fieldsign: using default parms; new parmfile is:  field.parm\n");
    fpparm = fopen("field.parm","w");
    fprintf(fpparm,"MIN_DIFF = %3.2f\n",MIN_DIFF);
    fprintf(fpparm,"NON_MIRROR = %d\n",NON_MIRROR);
    fprintf(fpparm,"BETWEEN = %d\n",BETWEEN);
    fprintf(fpparm,"MIRROR = %d\n",MIRROR);
    fprintf(fpparm,"BORDER = %d\n",BORDER);
    fprintf(fpparm,"LEFT_HEMISPHERE = %d\n",LEFT_HEMISPHERE);
    fprintf(fpparm,"FLIPPED_HEMISPHERE = %d\n",FLIPPED_HEMISPHERE);
    fprintf(fpparm,"SIGMOID_SLOPE = %3.2f\n",SIGMOID_SLOPE);
    fclose(fpparm);
  }
  else {  /* read any parms in file; others leave default */
    printf("fieldsign: using **existing** parmfile:\n");
    while (fgets(line,80,fpparm) != NULL) { 
      if (sscanf(line,"%s %s %f",str,junk,&parm) == 3)  {
        if      (MATCH_STR("MIN_DIFF"))    min_diff = parm;
        else if (MATCH_STR("NON_MIRROR"))  non_mirror = (int)parm;
        else if (MATCH_STR("BETWEEN"))     between = (int)parm;
        else if (MATCH_STR("MIRROR"))      mirror = (int)parm;
        else if (MATCH_STR("BORDER"))      border = (int)parm;
        else if (MATCH_STR("LEFT_HEMISPHERE"))    left_hemisphere = (int)parm;
        else if (MATCH_STR("FLIPPED_HEMISPHERE")) flipped_hemisphere=(int)parm;
        else if (MATCH_STR("SIGMOID_SLOPE"))      sigmoid_slope = parm;
        else printf("fieldsign: don't know this parm: %s  ...ignored\n",str);
      }
      else {
        printf("fieldsign: ### couldn't parse this field.parm line: %s",line);
        printf("  ...quitting\n");
        exit(1);
      }
    }  /* lines */
    fclose(fpparm);
  } /* done reading parms */

  /* swap fieldsign coloring for left hemisphere and/or flip hemisphere */
  if (left_hemisphere) {   /* assumes theta -90 => 90 for both RH and LH */
    tmp = mirror;
    mirror = non_mirror;
    non_mirror = tmp;
  }
  if (flipped_hemisphere) {
    tmp = mirror;
    mirror = non_mirror;
    non_mirror = tmp;
  }

  /* open infiles */
  sprintf(str,"%s.radius.%dx%d",argv[1],xsize,ysize);
  fpradius = fopen(str,"r");
  if (fpradius==NULL) {
    printf("fieldsign: ### couldn't find radius data file: %s\n",str);
    exit(1);
  }
  sprintf(str,"%s.theta.%dx%d",argv[1],xsize,ysize);
  fptheta = fopen(str,"r");
  if (fptheta==NULL) {
    printf("fieldsign: ### couldn't find theta data file: %s\n",str);
    exit(1);
  }

  /* malloc */
  radius_array = (float **)malloc(ysize*sizeof(float *)); 
  for (i=0; i<ysize; i++)
    radius_array[i] = (float *)malloc(xsize*sizeof(float));
  theta_array = (float **)malloc(ysize*sizeof(float *));  
  for (i=0; i<ysize; i++)
    theta_array[i] = (float *)malloc(xsize*sizeof(float));

  /* read data in */
  for (yi = 0; yi < ysize; yi++)
  for (xi = 0; xi < xsize; xi++)  {
    fscanf(fpradius,"%d %d %f",&x,&y,&radius_deg);  /* OK if shuffled */
    radius_array[y][x] = radius_deg*CONVERT;
    fscanf(fptheta,"%d %d %f",&x,&y,&theta_deg);
    theta_array[y][x] = theta_deg*CONVERT;
  }
  fclose(fpradius);
  fclose(fptheta);

  /* open outfile */
  sprintf(str,"%s.sign.%dx%d",argv[1],xsize,ysize);
  printf("fieldsign: outfile: %s\n",str);
  fpsign = fopen(str,"w");

  /* use two infiles to find local visual field sign; write output rasterfile*/
  for (y=0; y<ysize; y++)
  for (x=0; x<xsize; x++) {
    if (x<offset || y<offset || x>=xsize-offset || y>=ysize-offset) {
      color = border;        
      fprintf(fpsign,"%d\t%d\t%d\n",x,y,color);
    }
    else {  /* not border */

      /* find gradients for radius and theta arrays */
      dx_radius = radius_array[y][x+offset] - radius_array[y][x-offset];
      dy_radius = radius_array[y+offset][x] - radius_array[y-offset][x];
      radius_grad_angle = atan2((double)dy_radius,(double)dx_radius);
      dx_theta = theta_array[y][x+offset] - theta_array[y][x-offset];
      dy_theta = theta_array[y+offset][x] - theta_array[y-offset][x];
      theta_grad_angle = atan2((double)dy_theta,(double)dx_theta);

      /* catch wraparound for gradient direction angle subtraction */
      if (radius_grad_angle - theta_grad_angle > M_PI)
        theta_grad_angle += 2.0*M_PI;
      if (radius_grad_angle - theta_grad_angle < -M_PI)
        radius_grad_angle += 2.0*M_PI;

      /* find local field sense */
      sense_angle = radius_grad_angle - theta_grad_angle;

      /* coloring options */
      if (color_option == BINARY) {
        if (sense_angle > (-M_PI+min_diff) && sense_angle < -min_diff)
          color = mirror;
        else if (sense_angle > min_diff && sense_angle < (M_PI-min_diff))
          color = non_mirror;
        else
          color = between;
      }
      else if (color_option == SIGMOID) {   /* mult 3 sigs too slow */
        if (sense_angle >= -M_PI && sense_angle < -M_PI_2)       /* left 1/4*/
          sig = tanh((sense_angle+M_PI)*sigmoid_slope);
        else if (sense_angle >= -M_PI_2 && sense_angle < M_PI_2) /* mid 1/2*/
          sig = -tanh(sense_angle*sigmoid_slope); /* NB:small th is mirror */
        else                                                     /* right 1/4*/
          sig = tanh((sense_angle-M_PI)*sigmoid_slope);
        color = (int)((float)non_mirror +
                               0.5*(1.0+sig)*((float)(mirror-non_mirror)));
      }
      else if (color_option == GRAYSCALE)
        color = (int)((sense_angle + M_PI) * (255.0/(2.0 * M_PI)));
      fprintf(fpsign,"%d\t%d\t%d\n",x,y,color);

    }
  }
  fclose(fpsign);
  exit(0);
}

/*** begin functions */
void print_help(void)
{
printf("\
#############################################################################\n\
        fieldsign -- calculate visual fieldsign (mirror vs. non-mirror)\n\
\n\
                       written by Marty Sereno\n\
                  University of California, San Diego\n\
#############################################################################\n\
     Reads two ASCII rasterfiles of r and theta data in degrees.  Must\n\
  have file dimensions in filename as well as command line parms.\n\
\n\
    rootfilename.radius.30x25     format: x  y  radius(deg)\n\
    rootfilename.theta.30x25      format: x  y  theta(deg)\n\
\n\
  Calculates local visual field sign at each point--i.e., mirror-image\n\
  (like V1) or non- mirror-image (like V2)--using <offset> pixels\n\
  in N/S/E/W direction to calculate gradients for a little smoothing.\n\
  An offset=1 means gradients calculated across 3 pixels, offset=2\n\
  across 5 pixels, etc.  MIRROR and NON_MIRROR will be correct for\n\
  right-hemisphere/left-hemifield data entered on the digitizing\n\
  sphere coordinate system with theta ranging from -90 => 90 (i.e.,\n\
  with a mirror-reversal).  Set LEFT_HEMISPHERE = 1 to get correct\n\
  fieldsign for *left* cortical hemisphere data entered with theta\n\
  from -90 => 90.  Set FLIPPED_HEMISPHERE if arrows was used to\n\
  flip hemisphere.\n\
\n\
  ********************************************************************\n\
      ==> To get correct mirror/non-mirror colors, *don't* flip \n\
         *hemifield* when making *.out file with arrows. <==\n\
  ********************************************************************\n\
\n\
    Also reads parm file (field.parm) file containing various settings\n\
  (e.g., output colors, hemisphere, flip); if not there writes a default\n\
  version of the file, which can then be edited.  Ignores properly\n\
  formatted parameters in parm file that it doesn't recognize.  If a\n\
  parm appears more than once, the last instance determines its value.\n\
  Parms omitted from file are set to default values.  Default parm file:\n\
\n\
MIN_DIFF = 0.00\n\
NON_MIRROR = 100\n\
BETWEEN = 175\n\
MIRROR = 200\n\
BORDER = 250\n\
LEFT_HEMISPHERE = 0\n\
FLIPPED_HEMISPHERE = 0\n\
SIGMOID_SLOPE = 4.00\n\
\n\
  Writes single output ASCII rasterfile:\n\
\n\
    rootfilename.sign.30x25     format: x  y  color\n\
\n\
  where color is a grayscale number (0-255), with three alternate\n\
  color options:\n\
\n\
  colopt=0:   0-255 grayscale levels, where black to gray is\n\
              NON_MIRROR, and and gray to white is MIRROR\n\
\n\
  colopt=1:   three-level grayscale (MIRROR/NON_MIRROR/BETWEEN) with\n\
              amount of angle dumped into BETWEEN set by MIN_DIFF\n\
\n\
  colopt=2:   sigmoid grayscale, where a tanh() function is used to\n\
              clamp colors to MIRROR and NON_MIRROR, with a smooth\n\
              transition between.  SIGMOID_SLOPE = 1.0 is approximately\n\
              a sine, so use SIGMOID_SLOPE = 4.0 to get a transition\n\
              over a narrower range.\n\
\n\
  Writes <offset> wide BORDER-colored pixels at edge of map (no\n\
  gradient calculated).\n\
#############################################################################\n\
\n\
");
}

