/*#############################################################################
 interpolate.c: interpolate sparse x-y vector data onto grid -- Marty Sereno
###############################################################################
91/07/03 -- first hack
91/09/03 -- rotate RFs; output in deg; real ctx mm datafile parm
92/01/21 -- xy outfile to plot points
92/03/10 -- check rounding (bad array dim)
92/12/08 -- remove transforms duplicated in arrows.c
93/01/26 -- fix mesh sizing
93/03/16 -- write x,y,z and x,y files for GMT, epsilon is parm
93/03/20 -- rectangular grid elements
93/04/09 -- write x,y,z lesions file (if *-les there)
94/06/01 -- write file for GMT pstext
14/10/08 -- format, ANSI-ify
17/09/25 -- option to print help
#############################################################################*/
/* compile: gcc -D{Linux,Darwin} interpolate.c -lm -o interpolate.c */

#include <stdio.h>
#include <stdlib.h>
#include <math.h>
/*#include <jv.h>*/      /* NeXT fixed transcendental mathlib :-} */
#include <string.h>

/* default parms: adjust in parm file */
#define  X_GRIDSIZE     30   /* grid elements (number of vertices) */
#define  SMOOTHING_RAD  0.3  /* ignore pnts beyond this (frac of x_gridsize */
#define  SMOOTHNESS     1.2  /* relative effect of nearer data points */
#define  DATA_MARG_FRAC 0.02 /* fraction of total grid beyond max x data */
#define  EPSILON        0.1  /* weighting Gaussian height (smaller=>higher) */
#define  Y_TO_X_STIFFNESS  1.0  /* stretch data so stiffer by this in y dir */

/* limits, etc */
#define  MAX_POINTS     10000     /* should malloc */
#define  MAX_NAMESIZE      20     /* ditto */
#define  CONVERT (2.0*M_PI/360.0) /* degrees times this ==> radians */

/* true for exact match between S and str */
#define  MATCH_STR(S)   (strncmp(str,S,strlen(S)) == 0)
#define  SQR(x)         ((x)*(x))

/* function declarations */
int main(int argc, char **argv);
void print_help(void);


/*** begin main */
int main(int argc, char **argv)
{
  char   line[200], str[200], junk[5];

  int    i, gridcnt, gridnum, helpflag=0;
  int    x_grid, y_grid;
  int    numpts, num_outside;
  int    x_gridsize, y_gridsize;

  float  parm;
  float  x, y;
  float  radius_deg, theta_deg, length_deg, width_deg, angle_deg, dia_deg;
  float  ymax=0.0, xmax=0.0, ymin=0.0, xmin=0.0, xsize, ysize;
  float  smoothing_rad, smoothness, epsilon, y_to_x_stiffness;
  float  x_incr, y_incr;
  float  total_xsize, total_ysize;
  float  data_marg, y_cent_data_marg, data_marg_frac;
  float  xdiff, ydiff;
  float  sq_dist, weight, sum_weights, sq_cutoff_dist;
  float  ra_weighted_sum, th_weighted_sum, di_weighted_sum;

  FILE   *fopen(), *fpparm, *fpdata, *fpradius, *fptheta, *fpdia, *fpxy;
  FILE   *fpxytext;

  typedef struct rf_type_ {
    char   name[MAX_NAMESIZE];
    float  x;
    float  y;
    float  radius;
    float  theta;
    float  dia;
  } rf_type;
  rf_type   rf[MAX_POINTS];

  if (argc==2) {
    strcpy(str,argv[1]);
    if ((MATCH_STR("-h")) || (MATCH_STR("-help"))) helpflag=1;
  }
  if (argc!=2 || helpflag)  {
    printf("\n");
    printf("usage: interpolate <rootfilename>\n");
    printf("       interpolate -h,-help\n");
    printf("\n");
    printf("parms and    X_GRIDSIZE = 30        EPSILON = 0.1\n");
    printf("defaults:    SMOOTHING_RAD = 0.3    DATA_MARG_FRAC = 0.02\n");
    printf("             SMOOTHNESS = 1.2       Y_TO_X_STIFFNESS = 1.0\n");
    printf("[version 14]\n");
    printf("\n");
    if (helpflag) {
      strcpy(str,argv[1]);
      if ((MATCH_STR("-h")) || (MATCH_STR("-help"))) { print_help(); }
    }
    exit(0);
  }

  sprintf(str,"%s.str",argv[1]);
  fpdata = fopen(str,"r");
  if (fpdata==NULL)  {
    sprintf(str,"%s.out",argv[1]);
    fpdata = fopen(str,"r");
  }
  if (fpdata==NULL)  {
    printf("interpolate: ### couldn't find transformed file (%s.str, %s.out)\n",
           argv[1],argv[1]);
    exit(1);
  }
  printf("interpolate: using datafile: %s\n",str);

  /* load default parms */
  x_gridsize = X_GRIDSIZE;
  smoothing_rad = SMOOTHING_RAD;
  smoothness = SMOOTHNESS;
  epsilon = EPSILON;
  data_marg_frac = DATA_MARG_FRAC;
  y_to_x_stiffness = Y_TO_X_STIFFNESS;

  /* try to open parms file */
  fpparm = fopen("int.parm","r");
  if (fpparm==NULL) {
    printf("interpolate: using default parms; new parmfile is:  int.parm\n");
    fpparm = fopen("int.parm","w");
    fprintf(fpparm,"X_GRIDSIZE = %d\n",X_GRIDSIZE);
    fprintf(fpparm,"SMOOTHING_RAD = %3.2f\n",SMOOTHING_RAD);
    fprintf(fpparm,"SMOOTHNESS = %3.2f\n",SMOOTHNESS);
    fprintf(fpparm,"EPSILON = %3.4f\n",EPSILON);
    fprintf(fpparm,"DATA_MARG_FRAC = %3.2f\n",DATA_MARG_FRAC);
    fprintf(fpparm,"Y_TO_X_STIFFNESS = %3.2f\n",Y_TO_X_STIFFNESS);
    fclose(fpparm);
  }
  else {  /* read any parms in file, others leave default */
    printf("interpolate: using **existing** parmfile:\n");
    while (fgets(line,80,fpparm)!=NULL) { 
      if (sscanf(line,"%s %s %f",str,junk,&parm) == 3)  {
        if      (MATCH_STR("X_GRIDSIZE"))       x_gridsize = (int)parm;
        else if (MATCH_STR("SMOOTHING_RAD"))    smoothing_rad = parm;
        else if (MATCH_STR("SMOOTHNESS"))       smoothness = parm;
        else if (MATCH_STR("EPSILON"))          epsilon = parm;
        else if (MATCH_STR("DATA_MARG_FRAC"))   data_marg_frac = parm;
        else if (MATCH_STR("Y_TO_X_STIFFNESS")) y_to_x_stiffness = parm;
        else printf("interpolate: don't know parm: %s  ...ignored\n",str);
        }
      else {
        printf("interpolate: ### couldn't parse line in int.parm:  %s",line);
        printf("  ...quitting\n");
        exit(1);
      }
    }
    fclose(fpparm);
  }

  /* get datafile line; read 8 args data (won't read oldstyle 6-arg lines) */
  i = numpts = 0;
  while (fgets(line,80,fpdata) != NULL) { 
    if (sscanf(line,"%s %f %f %f %f %f %f %f",rf[i].name,&x,&y,
               &radius_deg,&theta_deg,&length_deg,&width_deg,&angle_deg) == 8)
      dia_deg = (length_deg + width_deg)/2.0;
    else {
      printf("interpolate: couldn't parse this line in data file:\n");
      printf("   %s",line);
      continue;
    }
    /***DEBUG: if (x>10.0||x<5.0||y>10.0||y<5.0) continue; *get pnts for test*/

    /* maxs,mins */ 
    if (i==0) { ymax=ymin=y; xmax=xmin=x; }
    if (x>xmax) xmax=x; if (y>ymax) ymax=y;
    if (x<xmin) xmin=x; if (y<ymin) ymin=y;

    /* store in struct (convert to radians) */
    rf[i].x = x;
    rf[i].y = y;
    rf[i].radius = radius_deg*CONVERT;
    rf[i].theta = theta_deg*CONVERT;
    rf[i].dia = dia_deg*CONVERT;
    i++;
    if (i>=MAX_POINTS) {
      printf("interpolate: ### overflowed %d maxpts\n",MAX_POINTS); exit(1);}
  }  /*** for all lines in file */

  fclose(fpdata);
  numpts = i;
  ysize = ymax - ymin;
  xsize = xmax - xmin;
  printf("interpolate: x(min,max) = (%3.2f,%3.2f) mm\n",xmin,xmax);
  printf("interpolate: y(min,max) = (%3.2f,%3.2f) mm\n",ymin,ymax);
  /*** done with data file ***/

  /* get grid dimensions, margins; open outfiles (gridsize *not* zero-based) */
  data_marg = data_marg_frac*xsize;
  total_xsize = xsize + 2.0*data_marg;  /* in mm */
  total_ysize = ysize + 2.0*data_marg;
  x_incr = y_incr = total_xsize/(float)(x_gridsize-1);
  printf("interpolate: mm_per_pixel = %3.5f\n",x_incr);
  /***make diff gridpts in y (diff than stretch: less in sum but same exp)*/
  /***x_incr = total_xsize/(float)(x_gridsize-1);*/
  /***y_incr = x_incr*grid_y_to_x;*/
  y_gridsize = ceil(total_ysize/y_incr) + 1;
  y_cent_data_marg = ((float)(y_gridsize-1)*y_incr - total_ysize)/2.0;
  sprintf(str,"%s.radius.%dx%d",argv[1],x_gridsize,y_gridsize);
  fpradius = fopen(str,"w");
  printf("interpolate: eccen file: %s\n",str);
  sprintf(str,"%s.theta.%dx%d",argv[1],x_gridsize,y_gridsize);
  fptheta = fopen(str,"w");
  printf("interpolate: angle file: %s\n",str);
  sprintf(str,"%s.dia.%dx%d",argv[1],x_gridsize,y_gridsize);
  fpdia = fopen(str,"w");
  printf("interpolate: diameter file: %s\n",str);

  /* smooth radius and theta data; write files */
  num_outside = 0;
  gridcnt = x_gridsize * y_gridsize;
  sq_cutoff_dist = SQR(smoothing_rad * ((xsize+ysize)/2.0));  /* cutoff */

  for (y_grid=0; y_grid<y_gridsize; y_grid++)
  for (x_grid=0; x_grid<x_gridsize; x_grid++) {  /* for all grid points */
    sum_weights = ra_weighted_sum = th_weighted_sum = di_weighted_sum = 0.0;
    for (i=0; i<numpts; i++) {  /* go though data points */
      if (rf[i].radius<0.0) continue;  /* skip no-response points */

      /* get distance */
      xdiff = rf[i].x - (x_grid*x_incr - data_marg + xmin);
      ydiff = rf[i].y - (y_grid*y_incr - data_marg + ymin -y_cent_data_marg);

      /* adjust relative effect of points in y and x direction */
      sq_dist = SQR(xdiff) + SQR(ydiff/y_to_x_stiffness);
      if (sq_dist>sq_cutoff_dist) continue;

      /* ctx-mm-based sq_dist mixed with arbitrary smoothness,epsilon */
      weight = (exp(-smoothness*sq_dist))/(sq_dist + epsilon);  /*exp falloff*/
      sum_weights += weight;
      ra_weighted_sum += weight*rf[i].radius;
      th_weighted_sum += weight*rf[i].theta;
      di_weighted_sum += weight*rf[i].dia;
    }  /* for all data points */
    if (sum_weights==0.0) { sum_weights=1.0; num_outside++; }

    /* get weighted sum and convert back to degrees */
    fprintf(fpradius,"%d\t%d\t%3.3f\n",
            x_grid,y_grid,(ra_weighted_sum/sum_weights)/CONVERT);
    fprintf(fptheta,"%d\t%d\t%3.3f\n",
            x_grid,y_grid,(th_weighted_sum/sum_weights)/CONVERT);
    fprintf(fpdia,"%d\t%d\t%3.3f\n",
            x_grid,y_grid,(di_weighted_sum/sum_weights)/CONVERT);
    gridnum = y_grid*x_gridsize + x_grid;
    if (!(gridnum%(gridcnt/10))) {
      printf("....%d%%",(100*gridnum)/gridcnt); fflush(stdout);}
  }  /* grid points within smoothing rad */
  printf("\ninterpolate: %d gridpnts w/no data in smoothing rad\n",num_outside);
  fclose(fpradius);
  fclose(fptheta);
  fclose(fpdia);

  /* write clean file cortical xy coords, xytext in smoothing-pixel units */
  sprintf(str,"%s.xy.%dx%d",argv[1],x_gridsize,y_gridsize);
  fpxy = fopen(str,"w");
  printf("interpolate: data xy file: %s\n",str);
  sprintf(str,"%s.xytext.%dx%d",argv[1],x_gridsize,y_gridsize);
  fpxytext = fopen(str,"w");
  printf("interpolate: data xytext file: %s\n",str);

  /* N.B.: x_gridsize not zero-based (n pixels are n-1 pixels long) */
  for (i=0; i<numpts; i++) {
    x = (rf[i].x + data_marg - xmin)/total_xsize * (float)(x_gridsize-1);
    y = (rf[i].y + data_marg - ymin + y_cent_data_marg)/
                 (total_ysize + 2.0*y_cent_data_marg) * (float)(y_gridsize-1);
    fprintf(fpxy,"%f\t%f\n",x,y);
    fprintf(fpxytext,"%f\t%f\t%s\n",x,y,rf[i].name);
  }
  fclose(fpxy);
  fclose(fpxytext);

  /* write clean file cntl lesion xy coords in smoothing-pixel units if file */
  sprintf(str,"%s-les.out",argv[1]);
  fpdata = fopen(str,"r");
  if (fpdata == NULL)
    printf("interpolate: no transformed lesion file: %s\n",str);
  else  {
    printf("interpolate: using lesion file: %s\n",str);
    sprintf(str,"%s.les-xy.%dx%d",argv[1],x_gridsize,y_gridsize);
    fpxy = fopen(str,"w");
    printf("interpolate: lesion xy file: %s\n",str);
    while (fgets(line,80,fpdata) != NULL) { 
      if (sscanf(line,"%s %f %f %f %f %f %f %f",rf[i].name,&x,&y,
          &radius_deg,&theta_deg,&length_deg,&width_deg,&angle_deg) == 8) {
        x = (x + data_marg - xmin)/total_xsize * (float)(x_gridsize-1);
        y = (y + data_marg - ymin + y_cent_data_marg)/
                  (total_ysize + 2.0*y_cent_data_marg) * (float)(y_gridsize-1);
        fprintf(fpxy,"%f\t%f\n",x,y);
      }
      else  {
        printf("interpolate: couldn't parse this line in lesion file:\n");
        printf("   %s",line);
      }
    }
  }
  fclose(fpxy);
  exit(0);
}

/*** begin functions */
void print_help()
{
printf("\
#############################################################################\n\
           interpolate -- interpolate sparse x-y vector data onto grid\n\
\n\
                       written by Marty Sereno\n\
                  University of California, San Diego\n\
#############################################################################\n\
    Reads arrows.c format data file (ASCII table) of penetration\n\
  coordinates and RF params with data sparsely distributed in the\n\
  x-y plane.  Looks first for output of stretch (rootfilename.str)\n\
  and then output of arrows (rootfilename.out).  Writes 3 ASCII\n\
  files (r,theta,dia) of interpolated data on a regular grid with\n\
  x size set in parm file. Uses distance-weighted smoothing with\n\
  an adjustable Gaussian:\n\
\n\
      weight = (exp(-smoothness*sq_dist))/(sq_dist + epsilon);\n\
\n\
  The z value at a grid point is the weighted sum of the z values\n\
  of the surrounding data points scaled by the sum of the weights\n\
  used for that grid point.\n\
\n\
  Also writes ASCII outfile of cortex coords in smoothing-pixel units\n\
  so that location of the data points can be plotted on top of the\n\
  contours.  If there is a transformed lesion file (rootfilename-les.out),\n\
  writes out ASCII file of lesions coords in smoothing-pixel units so\n\
  that location of the lesions can also be plotted on top of the\n\
  contours with bigger points to distinguish them from the regular\n\
  recording points (since the lesions file serves as control points\n\
  for stretch.c, we don't need a stretched lesion file...).\n\
\n\
Infiles:\n\
  rootfilename.out              format: 8 args ASCII table (see below)\n\
  [rootfilename-les.out]        format: 8 args ASCII table (see below)\n\
Outfiles (ASCII):\n\
  rootfilename.radius.30x25     format: x  y  radius(deg)\n\
  rootfilename.theta.30x25      format: x  y  theta(deg)\n\
  rootfilename.dia.30x25        format: x  y  dia(deg)\n\
  rootfilename.xy.30x25         format: x  y\n\
  rootfilename.xytext.30x25     format: x  y  pntname\n\
  [rootfilename.les-xy.30x25]   format: x  y\n\
\n\
  Input data files are ASCII tables.  Format for data line in input\n\
  file (also read by arrows.c, rfplots.c, and stretch.c) is 8 args\n\
  (won't read old 6 arg lines w/just dia):\n\
\n\
pointname  point-x  point-y  RF-r  RF-theta  RF-length  RF-width  RF-angle\n\
\n\
  RF-r (distance from center of gaze to receptive field center)\n\
  given in 0-90+ degrees and RF-theta (angle of center-of-gaze-to-RF\n\
  line) in -90 to +90 degrees, RF-dia (or RF-length, RF-width) given\n\
  in positive degrees and RF-angle (measured relative to the long\n\
  axis of the receptive field) given in -90 to +90 degrees.  There\n\
  can be up to MAX_NAMESIZE chars in point name.  Ignores no response\n\
  points (RF-radius = -1).  Example datafile:\n\
\n\
12F	5.23	9.25	24	-25	7.5	4.5	50\n\
254-L	2.31	5.36	37	50	5.5	3.5	-60\n\
312G	10.55	4.46	10	-10	7.5	5	-40\n\
571	3.24	7.35	12	57	14	9	-27\n\
601B	6.01	10.05	17	-9	10	8	-5\n\
25NR	3.45	12.2	-1	0	0	0	0\n\
\n\
  Also reads parm file (int.parm) file containing various settings\n\
  (smoothing parameters, gridsize); if not there writes a default\n\
  version of the file, which can then be edited.  Ignores properly\n\
  formatted parameters in parm file that it doesn't recognize.  If\n\
  a parm appears more than once, the last instance determines its\n\
  value.  Default parm file:\n\
\n\
X_GRIDSIZE = 30\n\
SMOOTHING_RAD = 0.3\n\
SMOOTHNESS = 1.2\n\
EPSILON = 0.1\n\
DATA_MARG_FRAC = 0.02\n\
Y_TO_X_STIFFNESS = 1.0\n\
\n\
  X_GRIDSIZE is number of vertices in the x direction.  Number of\n\
  vertices in the y direction is calculated (for square grid) and\n\
  put in outfile names.\n\
\n\
  Turning up SMOOTHNESS makes a narrower Gaussian which emphasizes\n\
  the effect of nearby data points on the current grid value (makes\n\
  plot smoother locally).  High values of SMOOTHNESS result in\n\
  plateaus around the data points with sigmoid transitions.\n\
\n\
  EPSILON adjusts the height of the Gaussian weighting function.\n\
  Small values (0.001) make a very tall Gaussian that forces the\n\
  surface to go through the data points while a larger value (0.1)\n\
  results in a smoother surface.\n\
\n\
  Data points that are more than SMOOTHING_RAD away from the current\n\
  grid point are not considered (speed up, truncation errors).  The\n\
  number of *grid points* that are more than SMOOTHING_RAD away\n\
  from *any* data point are reported (their value is set to 1.0).\n\
\n\
  DATA_MARG_FRAC is fraction of X_GRIDSIZE to extend interpolated\n\
  mesh beyond data x and y max,min.\n\
\n\
  Y_TO_X_STIFFNESS is used to increase the effect of points in the\n\
  y direction:\n\
\n\
      sq_dist = SQR(xdiff) + SQR(ydiff/y_to_x_stiffness);\n\
#############################################################################\n\
\n\
");
}
