/*#############################################################################
 rfplots.c: make PostScript receptive fields plot -- Marty Sereno
###############################################################################
90/08/17 -- first hack
90/08/19 -- rf connec lines
90/08/28 -- reads ellipse data, parm file, dashed lines
90/09/02 -- flat correction
91/07/23 -- rotate RF coord syst
91/09/09 -- fix parmfile name; ignore parm in datafile; better legend
91/10/07 -- ignore phi>180; put rf centers in w/lines+rfs
92/05/19 -- left/right hemifield
92/06/09 -- filled RFs
92/09/08 -- line break bug
93/02/23 -- smaller PostScript output w/defs
93/03/04 -- vector transforms; fix RF-angle w/ROTATE_ON_CG
93/05/13 -- stringwidth for better left hemifield labels
93/09/20 -- extra dup bug(!); add RANDOM_COLOR, HEMIFIELD_FRONT flags
93/09/23 -- LEGEND_FLAG; override connec+pnt line thick variation w/filled RFs
94/05/07 -- explicit setting of individual RF/pnt/line parms in datafile
94/06/24 -- make RF label angle, stalksize adjustable, always omit B.S. label
94/08/11 -- random RF label angle for dense plots, quadrant flag
14/10/08 -- format, ANSI-ify
17/09/25 -- option to print help
20/08/25 -- embed cmdline/parmfile as comments in ps outfile
 TODO: better point icons
#############################################################################*/
/* compile: gcc -D{Linux,Darwin} rfplots.c -lm -o rfplots */

#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <string.h>

#ifdef IRIX
#  include <values.h>
#endif

#ifdef Darwin
#  include <i386/limits.h>    /* INT_MAX=2147483647 */
#  define MAXINT INT_MAX
#endif

/* default parms; adjust to taste in parm file */
#define  RF_LINE      0.4   /* RF lines */
#define  PNT_LINE     0.4   /* point lines */
#define  CONNEC_LINE  0.4   /* line connecting RFs */
#define  PNT_DIA      4.0   /* dia of rf points */
#define  HF_LINE      0.2   /* hemifield border and HM and VM lines */
#define  RF_FONTSIZE  9.0   /* for RF numbers */
#define  RF_LABEL_ANGLE 30.0 /* for RF numbers */
#define  RAND_WIDTH    0.0  /* width of random dist of RF_LABEL_ANGLE */
#define  STALK_RATIO   0.8  /* fraction of fontsize for stalk length */
#define  VF_FONTSIZE 12.0   /* for hemifield labels */
#define  MAX_ECC     45.0   /* max ecc hemifield; also 75,60,30,24,15,12,9 */
#define  FLAT_CORRECT   1   /* flag to turn on flat-correction */
#define  MARG        30.0   /* size of margin */
#define  MOVE_ON_HM    0.0  /* rotate around 3-D cartesian y-axis */
#define  MOVE_ON_VM    0.0  /* rotate around 3-D cartesian x-axis */
#define  ROTATE_ON_CG  0.0  /* rotate around 3-D cartesian z-axis */
#define  LEFT_HEMIFIELD  0  /* set flag to change to left hemifield */
#define  RF_WHITENESS  -1.0 /* set receptive field fill color (-1; 0-1) */
#define  RANDOM_COLOR    0  /* set to randomly color RFs (overrides lt/dk) */
#define  HEMIFIELD_FRONT 0  /* set to draw hemifield in front of RFs */
#define  LEGEND_FLAG     1  /* unset flag to get rid of legend */
#define  QUADRANT_FLAG   0  /* 0=both, 1=upper, -1=lower */

/* limits, etc */
#define  TOTAL_RFS       5000 
#define  MAX_NAMESIZE     11     /* max num name characters + 1 */
#define  LEGEND_FONT      12.0
#define  BS_COLOR         0.20   /* 0.0 (black) to 1.0 (white) */
#define  TOWARD_HOR       0.75   /* move label line end to hor (1=labelangle)*/
#define  DASH1            3.0    /* default dash length for /light */
#define  DASH2            2.0    /* default dash space for /light */

#define  TOTAL_PYSIZE     792.0   /* (A4=842.0) page ysize in points */
#define  TOTAL_PXSIZE     612.0   /* (A4=595.0) page xsize in points */
#define  CONVERT (2.0*M_PI/360.0)  /* degrees times this ==> radians */
#define  PNTS     0
#define  RFS      1
#define  NONE    -1
#define  BLACK      0
#define  DARK       1 
#define  LIGHT      2
#define  FILLED       1
#define  UNFILLED     0
#define  LOWER_QUAD    -1
#define  BOTH_QUAD      0
#define  UPPER_QUAD     1

/* true for exact match between S and str */
#define  MATCH_STR(S)    (strncmp(str,S,strlen(S)) == 0)
#define  RANDOM_FLOAT(L,H)  ((L)+((float)random()/(float)MAXINT)*((H)-(L)))

/* function declarations */
int main(int argc, char **argv);
void ps_header(FILE *fpout);
void ps_parms(FILE *fpout, int argc, char **argv);
void ps_hemifield(FILE *fpout, float center_x, float center_y, float max_rad,
                  float max_ecc, float hf_line, float radians_to_pts,
                  float vf_fontsize, int left_hemifield, int quadrant_flag);
void ps_rf(FILE *fpout, float cent_x, float cent_y, float r, float theta,
           float halflen, float halfwi, float angle, float thick,
           float whiteness, int flatflg);
void ps_rflabel(FILE *fpout, float x, float y, float r, float theta,
                float h_len, float h_wi, float angle, float thick,
                char *str, float font, int flatflg, int left_hemi,
                float label_angle_deg, float stalk_ratio);
void ps_text(FILE *fpout, char *str, float x, float y, float fontsize);
void ps_pnt(FILE *fpout, float center_x, float center_y, float dia,
            float thickness, int filled_flag, float filled_color);
void ps_rline(
  FILE *fpout, float x, float y, float dx, float dy, float thickness);
void ps_line(
  FILE *fpout, float x1, float y1, float x2, float y2, float thickness);
void ps_isoecc(FILE *fpout, float centx, float centy,float radius,
               float thickness, int left_hemifield_flag, int quadrant_flag);
void rotate_x(float r[3], float angle);
void rotate_y(float r[3], float angle);
void rotate_z(float r[3], float angle);
void spherpolar_to_cartesian(float r[3]);
void cartesian_to_spherpolar(float r[3]);
void print_help(void);


/*** begin main */
int main(int argc, char **argv)
{
  char   line[200], str[200], cjunk[5];

  int    i, numflag, lineflag, plottype, flat_correct, left_hemifield;
  int    colorflag, numpts, random_color, hemifield_front, quadrant_flag;
  int    filled_flag=0, lastfilled_flag=0, legend_flag, numargs;

  float  x, y, radius_deg, radius, theta_deg, theta, dia_deg;
  float  length_deg, width_deg, angle_deg;
  float  half_length, half_width, rf_whiteness, tmp_rf_white;
  float  rf_line, pnt_line, connec_line, pnt_dia, hf_line, max_ecc;
  float  rf_label_angle, rand_width, stalk_ratio, arg1, arg2, parm;
  float  max_rad, radians_to_pts, marg, pxsize, pysize;
  float  rf_fontsize, vf_fontsize;
  float  tmp_rf_line=0.0, tmp_connec_line=0.0;
  float  tmp_pnt_line=0.0, tmp_label_line=0.0, tmp_pnt_fill=0.0;
  float  tmp_pnt_dia=0.0, tmp_rf_label_angle=0.0;
  float  rfline, connecline, pntline, labelline, dash1, dash2;
  float  rffill, pntfill;
  float  center_x, center_y, starty, startx;
  float  laststartx=0.0, laststarty=0.0;
  float  lasttmp_pnt_dia=0.0, lasttmp_pnt_line=0.0, lasttmp_pnt_fill=0.0;
  float  move_on_hm, move_on_vm, rotate_on_cg, tmp, fjunk;
  float  r[3];

  FILE   *fpparm, *fpdata, *fpout;

  typedef struct rf_type_ {
    char   name[MAX_NAMESIZE];
    float  x;            /* save this to see if B.S. (9999.0) flag */
    float  radius_deg;   /* PostScript */
    float  theta_deg;    /* PostScript */
    float  length;
    float  width;
    float  angle_deg;    /* PostScript */
    int    colorflag;    /* flag */
    int    break_before; /* flag */
    float  rfline;       /* points */
    float  connecline;   /* points */
    float  pntline;      /* points */
    float  labelline;    /* points */
    float  dash1;        /* points */
    float  dash2;        /* points */
    float  rffill;       /* -1,0-1 */
    float  pntfill;      /* -1,0-1 */
  } rf_type;
  rf_type  rf[TOTAL_RFS];

  if (argc != 5)  {
    printf("\nuse: rfplots <datafile> <0=pts,1=rfs> <lineflag> <numflag>\n\n");
    printf("parms and    RF_LINE = 0.40          MARG = 30.00\n");
    printf("defaults:    PNT_LINE = 0.40         MOVE_ON_HM = 0.00\n");
    printf("             CONNEC_LINE = 0.40      MOVE_ON_VM = 0.00\n");
    printf("             PNT_DIA = 4.00          ROTATE_ON_CG = 0.00\n");
    printf("             HF_LINE = 0.20          LEFT_HEMIFIELD = 0\n");
    printf("             RF_FONTSIZE = 9.00      RF_WHITENESS = -1.0\n");
    printf("             MAX_ECC = 45.00         RANDOM_COLOR = 0\n");
    printf("             FLAT_CORRECT = 1        HEMIFIELD_FRONT = 0\n");
    printf("             RF_LABEL_ANGLE = 30.0   LEGEND_FLAG = 1\n");
    printf("             STALK_RATIO = 0.8       RAND_WIDTH = 0.0\n");
    printf("             VF_FONTSIZE = 12.00     QUADRANT_FLAG = 0\n");
    printf("\n");
    printf("optional datafile parms:      /break\n");
    printf("  (these apply to all RFs     /black /dark /light\n");
    printf("   that follow them and       /rfline <pnts>\n");
    printf("   override rfs.parm file)    /pntline <pnts>\n");
    printf("                              /labelline <pnts>\n");
    printf("                              /connecline <pnts> \n");
    printf("   [neg to unset dash =>]     /dash <pnts> <pnts>\n");
    printf("                              /rffill <0-1>\n");
    printf("                              /pntfill <0-1>\n");
    printf("[version 26]\n");
    printf("\n");
    if (argc==2) {
      strcpy(str,argv[1]);
      if ((MATCH_STR("-h")) || (MATCH_STR("-help"))) { print_help(); }
    }
    exit(0);
  }

  plottype = atoi(argv[2]);
  lineflag = atoi(argv[3]);
  numflag = atoi(argv[4]);
  if (plottype>1||plottype<0||lineflag>1||lineflag<0||numflag>1||numflag<0) {
    printf("rfplots: ### bad flag\n");
    exit(1);
  }
  fpdata = fopen(argv[1],"r");
  if (fpdata==NULL)  {
    printf("rfplots: ### couldn't find data file: %s\n",argv[1]);
    exit(1);
  }

  /* load default parms */
  rf_line = RF_LINE;
  pnt_line = PNT_LINE;
  connec_line = CONNEC_LINE;
  pnt_dia = PNT_DIA;
  hf_line = HF_LINE;
  rf_fontsize = RF_FONTSIZE;
  rf_label_angle = RF_LABEL_ANGLE;
  rand_width = RAND_WIDTH;
  stalk_ratio = STALK_RATIO;
  vf_fontsize = VF_FONTSIZE;
  max_ecc = MAX_ECC;
  flat_correct = FLAT_CORRECT;
  marg = MARG;
  move_on_hm = MOVE_ON_HM;
  move_on_vm = MOVE_ON_VM;
  rotate_on_cg = ROTATE_ON_CG;
  left_hemifield = 0;
  rf_whiteness = RF_WHITENESS;
  random_color = 0;
  hemifield_front = 0;
  legend_flag = 1;
  quadrant_flag = BOTH_QUAD;

  /* try to open parms file */
  fpparm = fopen("rfs.parm","r");
  if (fpparm==NULL) {
    printf("rfplots: using default parms; new parmfile is:  rfs.parm\n");
    fpparm = fopen("rfs.parm","w");
    fprintf(fpparm,"RF_LINE = %3.2f\n",RF_LINE);
    fprintf(fpparm,"PNT_LINE = %3.2f\n",PNT_LINE);
    fprintf(fpparm,"CONNEC_LINE = %3.2f\n",CONNEC_LINE);
    fprintf(fpparm,"PNT_DIA = %3.2f\n",PNT_DIA);
    fprintf(fpparm,"HF_LINE = %3.2f\n",HF_LINE);
    fprintf(fpparm,"RF_FONTSIZE = %3.2f\n",RF_FONTSIZE);
    fprintf(fpparm,"RF_LABEL_ANGLE = %3.2f\n",RF_LABEL_ANGLE);
    fprintf(fpparm,"RAND_WIDTH = %3.2f\n",RAND_WIDTH);
    fprintf(fpparm,"STALK_RATIO = %3.2f\n",STALK_RATIO);
    fprintf(fpparm,"VF_FONTSIZE = %3.2f\n",VF_FONTSIZE);
    fprintf(fpparm,"MAX_ECC = %3.2f\n",MAX_ECC);
    fprintf(fpparm,"FLAT_CORRECT = %d\n",FLAT_CORRECT);
    fprintf(fpparm,"MARG = %3.2f\n",MARG);
    fprintf(fpparm,"MOVE_ON_HM = %3.2f\n",MOVE_ON_HM);
    fprintf(fpparm,"MOVE_ON_VM = %3.2f\n",MOVE_ON_VM);
    fprintf(fpparm,"ROTATE_ON_CG = %3.2f\n",ROTATE_ON_CG);
    fprintf(fpparm,"LEFT_HEMIFIELD = %d\n",LEFT_HEMIFIELD);
    fprintf(fpparm,"RF_WHITENESS = %3.2f\n",RF_WHITENESS);
    fprintf(fpparm,"RANDOM_COLOR = %d\n",RANDOM_COLOR);
    fprintf(fpparm,"HEMIFIELD_FRONT = %d\n",HEMIFIELD_FRONT);
    fprintf(fpparm,"LEGEND_FLAG = %d\n",LEGEND_FLAG);
    fprintf(fpparm,"QUADRANT_FLAG = %d\n",QUADRANT_FLAG);
    fclose(fpparm);
  }
  else {  /* read any parms in file, others leave default */
    printf("rfplots: using **existing** parmfile:\n");
    while (fgets(line,80,fpparm) != NULL) { 
      if (sscanf(line,"%s %s %f",str,cjunk,&parm) == 3)  {
        if      (MATCH_STR("RF_LINE"))       rf_line = parm;
        else if (MATCH_STR("PNT_LINE"))      pnt_line = parm;
        else if (MATCH_STR("CONNEC_LINE"))   connec_line = parm;
        else if (MATCH_STR("PNT_DIA"))       pnt_dia = parm;
        else if (MATCH_STR("HF_LINE"))       hf_line = parm;
        else if (MATCH_STR("RF_FONTSIZE"))   rf_fontsize = parm;
        else if (MATCH_STR("RF_LABEL_ANGLE"))rf_label_angle = parm;
        else if (MATCH_STR("RAND_WIDTH"))    rand_width = parm;
        else if (MATCH_STR("STALK_RATIO"))   stalk_ratio = parm;
        else if (MATCH_STR("VF_FONTSIZE"))   vf_fontsize = parm;
        else if (MATCH_STR("MAX_ECC"))       max_ecc = parm;
        else if (MATCH_STR("FLAT_CORRECT"))  flat_correct = (int)parm;
        else if (MATCH_STR("MARG"))          marg = (int)parm;
        else if (MATCH_STR("MOVE_ON_HM"))    move_on_hm = parm;
        else if (MATCH_STR("MOVE_ON_VM"))    move_on_vm = parm;
        else if (MATCH_STR("ROTATE_ON_CG"))  rotate_on_cg = parm;
        else if (MATCH_STR("LEFT_HEMIFIELD")) left_hemifield = (int)parm;
        else if (MATCH_STR("RF_WHITENESS"))  rf_whiteness = parm;
        else if (MATCH_STR("RANDOM_COLOR"))  random_color = (int)parm;
        else if (MATCH_STR("HEMIFIELD_FRONT")) hemifield_front = (int)parm;
        else if (MATCH_STR("LEGEND_FLAG"))   legend_flag = (int)parm;
        else if (MATCH_STR("QUADRANT_FLAG")) quadrant_flag = (int)parm;
        else printf("rfplots: don't know this parm: %s  ...ignored\n",str);
     }
     else {
        printf("rfplots: ### couldn't parse this line in rfs.parm:  %s",line);
        printf("  ...quitting\n");
        exit(1);
      }
    }
    fclose(fpparm);
  }

  /* use parmfile to set defaults for explicit args that might not be set */
  for (i=0; i<TOTAL_RFS; i++)
    rf[i].break_before = 0;
  colorflag = DARK;   /* default symbolic color */
  rfline = rf_line;
  connecline = connec_line;
  pntline = pnt_line;
  labelline = rf_line;
  dash1 = -1.0;
  dash2 = -1.0;
  rffill = rf_whiteness;
  pntfill = rf_whiteness;
    
  i = numpts = 0;  /* some points might be no-rf flags */

  /* get line; skip comments,space; get commands; read 6 or 8 args data */
  while (fgets(line,80,fpdata) != NULL) { 
    if (line[0] == '#' || line[0] == ' ' || line[0] == '\n')  /* comment */
      continue;
    if (line[0] == '/') {  /* command */
      numargs = sscanf(line,"%s %f %f",str,&arg1,&arg2);
      if (MATCH_STR("/break") && numargs==1) {
        rf[i].break_before = 1;
        continue;
      }
      else if (MATCH_STR("/black") && numargs==1) {
        colorflag = BLACK;
        continue;
      }
      else  if (MATCH_STR("/dark") && numargs == 1)  {
        colorflag = DARK;
        continue;
      }
      else if (MATCH_STR("/light") && numargs == 1)  {
        colorflag = LIGHT;
        continue;
      }
      else if (MATCH_STR("/rfline") && numargs == 2)  {
        colorflag = NONE;
        rfline = arg1;
        continue;
      }
      else if (MATCH_STR("/connecline") && numargs == 2)  {
        colorflag = NONE;
        connecline = arg1;
        continue;
      }
      else if (MATCH_STR("/pntline") && numargs == 2)  {
        colorflag = NONE;
        pntline = arg1;
        continue;
      }
      else if (MATCH_STR("/labelline") && numargs == 2)  {
        colorflag = NONE;
        labelline = arg1;
        continue;
      }
      else if (MATCH_STR("/dash") && numargs == 3)  {
        colorflag = NONE;
        dash1 = arg1;
        dash2 = arg2;
        continue;
      }
      else if (MATCH_STR("/rffill") && numargs == 2)  {
        colorflag = NONE;
        rffill = arg1;
        continue;
      }
      else if (MATCH_STR("/pntfill") && numargs == 2)  {
        colorflag = NONE;
        pntfill = arg1;
        continue;
      }
      else {
        printf("rfplots: ### command: %s  not recognized  ...skipped\n",line);
        continue;
      }
    }

    /* save RF parms read last time from commands (or left default) */
    rf[i].colorflag = colorflag;
    rf[i].rfline = rfline;
    rf[i].connecline = connecline;
    rf[i].pntline = pntline;
    rf[i].labelline = labelline;
    rf[i].dash1 = dash1;
    rf[i].dash2 = dash2;
    rf[i].rffill = rffill;
    rf[i].pntfill = pntfill;

    /* else-if block; remember, try longest line first!! */
    /* check if an 8 arg dataline */
    if (sscanf(line,"%s %f %f %f %f %f %f %f",rf[i].name,&x,&y,
               &radius_deg,&theta_deg,&length_deg,&width_deg,&angle_deg) == 8)
      dia_deg = (width_deg + length_deg)/2.0;

    /* check if a 6 arg dataline */
    else if (sscanf(line,"%s %f %f %f %f %f",rf[i].name,&x,&y,
             &radius_deg,&theta_deg,&dia_deg) == 6)  {
      if (dia_deg == 0.0) dia_deg = 5.0;  /* 0dia -> standard */
      width_deg = length_deg = dia_deg;
      angle_deg = 0.0;
    }

    /* check if a 3 arg parmline; ignore and jump out */
    else if (sscanf(line,"%s %s %f",str,cjunk,&fjunk) == 3)
      continue;

    /* else error */
    else {
      printf("rfplots: ### couldn't parse this line in data file:\n");
      printf("   %s",line);
      continue;
    }

    /* check rf params */
    if (width_deg>length_deg)  {
      printf("rfplots: ### RF length,width reversed for point: %s",rf[i].name);
      tmp = width_deg;
      width_deg = length_deg;
      length_deg = tmp;
      printf("  ...fixed\n");
    }

    /* ignore radius (phi,latitude) over 180 */
    if (radius_deg > 180.0)  {
      printf("rfplots: ### radius over 180 deg--datafile line ignored\n");
      printf("   %s",line);
      continue;
    }

    /* transform to cartesian; rotate RF coords; transform back */
    /* NOTE: hemiphere "radius"= spher. polar phi, "theta"= spherpolar theta */
    /* NOTE: rotations are like: x-along-HM, y-along-VM, rotate-around-CG */
    /* signs: rot,vm -> no neg (cw = neg angle, pos x is down [to viewer] */
    radius = radius_deg*CONVERT;
    theta = theta_deg*CONVERT;
    if (radius>0.0)  {  /* don't transform no-response points */
      r[0] = 1.0;
      r[1] = theta;
      r[2] = radius;
      spherpolar_to_cartesian(r);
      rotate_y(r,-move_on_hm*CONVERT);
      rotate_x(r,move_on_vm*CONVERT);
      rotate_z(r,rotate_on_cg*CONVERT);
      cartesian_to_spherpolar(r);
      theta = r[1];
      radius = r[2];
    }

    /* retain negative theta's for angles between -PI/2 to -PI */
    if (theta > M_PI && theta < 3.0*M_PI_2)   theta -= 2.0*M_PI;

    /* convert back to deg for PostScript */
    radius_deg = radius/CONVERT;
    theta_deg = theta/CONVERT;

    /* save in struct */
    rf[i].x = x;     /* save this to tell if B.S. */
    rf[i].radius_deg = radius_deg;  /* PostScript */
    rf[i].theta_deg = theta_deg;   /* PostScript */
    rf[i].length = length_deg*CONVERT;
    rf[i].width = width_deg*CONVERT;
    rf[i].angle_deg = angle_deg;   /* PostScript */
    i++;
    if (i>=TOTAL_RFS) {
      printf("rfplots: ### overflowed %d maxpnts)\n",TOTAL_RFS);
      exit(1);
    }
  }
  fclose(fpdata);
  numpts = i;

  /* get working coords */
  pxsize = TOTAL_PXSIZE - (2.0*marg);
  pysize = TOTAL_PYSIZE - (2.0*marg);

  /* open outfile and write PostScript header */
  sprintf(str,"%s.rfs.ps",argv[1]);
  printf("rfplots: PostScript outfile:  %s\n",str);
  fpout = fopen(str,"w");
  ps_header(fpout);
  ps_parms(fpout,argc,argv);

  /*** write filename and legend at upper right or left */
  if (legend_flag)  {
    sprintf(str,"file:  %s",argv[1]);
    if (left_hemifield) x = marg + 0.05*pxsize;
    else                x = marg + 0.75*pxsize;
    y = marg + pysize;
    ps_text(fpout,str,x,y-0.0*LEGEND_FONT,LEGEND_FONT);
    if (plottype==RFS)  {
      if (flat_correct)
        ps_text(fpout,"flat-corrected",x,y-1.0*LEGEND_FONT,LEGEND_FONT);
      else
        ps_text(fpout,"not flat-corrected",x,y-1.0*LEGEND_FONT,LEGEND_FONT);
    }
    if (move_on_hm!=0.0 || move_on_vm!=0.0 || rotate_on_cg!=0.0) {
      sprintf(str,"RF transformations");
      ps_text(fpout,str,x,marg+pysize-2.0*LEGEND_FONT,LEGEND_FONT);
      if (left_hemifield) x = marg + 0.07*pxsize;
      else                x = marg + 0.77*pxsize;
      sprintf(str,"HM:  %d deg",(int)move_on_hm);
      ps_text(fpout,str,x,y-3.0*LEGEND_FONT,LEGEND_FONT);
      sprintf(str,"VM:  %d deg",(int)move_on_vm);
      ps_text(fpout,str,x,y-4.0*LEGEND_FONT,LEGEND_FONT);
      sprintf(str,"Rot:  %d deg",(int)rotate_on_cg);
      ps_text(fpout,str,x,y-5.0*LEGEND_FONT,LEGEND_FONT);
    } /* if legend */
  }

  /* get points of largest radius semicircle (VM goes to edge...) */
  max_rad = pysize/2.0;

  /* get center of hemifield */
  if (left_hemifield) center_x = marg + (pxsize - max_rad)/2.0 + max_rad;
  else                center_x = marg + (pxsize - max_rad)/2.0;
  center_y = marg + max_rad;

  /* get factor to convert angles(rad) -> points */
  radians_to_pts = max_rad/(max_ecc*CONVERT); /* set size(deg) hemifield */

  /*** default: draw hemifield before RFs */
  if (!hemifield_front)
    ps_hemifield(fpout,center_x,center_y,max_rad,max_ecc,hf_line,
                 radians_to_pts,vf_fontsize,left_hemifield,quadrant_flag);

  /* loop to calculate RFs */
  for (i=0; i<numpts; i++) {

    /* find center point for RF */
    radius = rf[i].radius_deg*CONVERT;
    theta = rf[i].theta_deg*CONVERT;
    startx = center_x + radius*radians_to_pts * cos((double)theta);
    if (left_hemifield)  /* flip RF centers */
      startx = center_x + radius*radians_to_pts * cos((double)(M_PI-theta));
    else
      startx = center_x + radius*radians_to_pts * cos((double)theta);
    starty = center_y + radius*radians_to_pts * sin((double)theta);
    if (i==0) { laststartx=startx; laststarty=starty; }

    /*** draw RFs (num'd) of right col (RF-angle: coord trans only does ctr)*/
    if (plottype==RFS && rf[i].radius_deg>0.0)  {
      half_width = (rf[i].width/2.0) * radians_to_pts;
      half_length = (rf[i].length/2.0) * radians_to_pts;
      if (left_hemifield) {  /* now flip flat-correct (theta), RF angle fix*/
        theta_deg = 180.0 - rf[i].theta_deg;
        angle_deg = 180.0 - rf[i].angle_deg + rotate_on_cg;
      }
      else  {
        theta_deg = rf[i].theta_deg;
        angle_deg = rf[i].angle_deg - rotate_on_cg;
      }
      if (rf[i].x == 9999.0) {  /* force B.S. fill; override variation */
        tmp_rf_white = BS_COLOR;
        rf[i].colorflag = DARK;
      }
      else if (random_color) {  /* force rand color, override variation */
        tmp_rf_white = RANDOM_FLOAT(0.05,0.9);
        rf[i].colorflag = DARK;
      }
      else
        tmp_rf_white = rf_whiteness;

      /* set RF line thickness/dash and RF color */
      if (rf[i].colorflag==BLACK) {
        tmp_label_line = rf_line;
        tmp_rf_white = 2.0*tmp_rf_white - 1.0; /* full:RF_WHITENESS=0.5 */
        if (tmp_rf_white<0.0 && tmp_rf_white>= -1.0)
          tmp_rf_white = 0.0;  /* clip but leave RF_WHITENESS=-1.0 clear */
        tmp_rf_line = 2.0*rf_line;
        if (lineflag && rf_whiteness>0.0)  /* override line variation */
          tmp_rf_line = rf_line;
      }
      if (rf[i].colorflag==DARK) {
        tmp_label_line = rf_line;
        tmp_rf_line = rf_line;
      }
      if (rf[i].colorflag==LIGHT) {
        tmp_label_line = rf_line;
        if (tmp_rf_white>0.0 && tmp_rf_white<1.0)
          tmp_rf_white = 0.9;  /* light always same unless white or clear */
        fprintf(fpout,"[%3.1f %3.1f] 0 d\n",DASH1,DASH2);  /* PS:setdash */
        tmp_rf_line = 0.5*rf_line;
        if (lineflag && rf_whiteness>0.0) {  /* override line variation */
          tmp_rf_line = rf_line;
          fprintf(fpout,"[] 0 d\n");  /* PS:setdash--unset dash */
        }
      }
      if (rf[i].colorflag==NONE) {  /* force explicit values */
        tmp_label_line = rf[i].labelline;
        tmp_rf_line = rf[i].rfline;
        tmp_rf_white = rf[i].rffill;
        if (rf[i].dash1 >= 0.0)   /* PS: setdash */
          fprintf(fpout,"[%3.1f %3.1f] 0 d\n",rf[i].dash1,rf[i].dash2);
        else
          fprintf(fpout,"[] 0 d\n");
      }

      ps_rf(fpout,startx,starty,rf[i].radius_deg,theta_deg,
             half_length,half_width,angle_deg,tmp_rf_line,
             tmp_rf_white,flat_correct);
      fprintf(fpout,"[] 0 d\n");  /* PS:setdash--always unset dash */
      if (numflag && rf[i].x != 9999.0)  {  /* labels */
        tmp_rf_label_angle = rf_label_angle;
        if (rand_width != 0.0)
          tmp_rf_label_angle += RANDOM_FLOAT(-rand_width/2.0,rand_width/2.0);
        ps_rflabel(fpout,startx,starty,rf[i].radius_deg,theta_deg,
                   half_length,half_width,angle_deg,tmp_label_line,
                   rf[i].name,rf_fontsize,flat_correct,left_hemifield,
                   tmp_rf_label_angle,stalk_ratio);
      }
    }  /* if RF */

    /*** draw lines if needed */ 
    if (lineflag)  {
      /* if NR after break, fix *next* point */
      if (rf[i].radius_deg<0.0 && rf[i].break_before)
        rf[i+1].break_before = 1;

      /* draw line if RFthere + nobreak + notfirst */
      if (rf[i].radius_deg>0.0 && !rf[i].break_before && i) {
        if (lineflag && rf_whiteness > 0.0 && plottype == RFS) /* override */
          rf[i].colorflag = DARK;
        if (rf[i].colorflag==BLACK)
          tmp_connec_line = 2.0*connec_line;
        if (rf[i].colorflag==DARK)
          tmp_connec_line = connec_line;
        if (rf[i].colorflag==LIGHT) {
          fprintf(fpout,"[%3.1f %3.1f] 0 d\n",DASH1,DASH2);  /* PS:setdash */
          tmp_connec_line = 0.5*connec_line;
        }
        if (rf[i].colorflag==NONE) {  /* force values */
          tmp_connec_line = rf[i].connecline;
          if (rf[i].dash1>=0.0)   /* PS: setdash */
            fprintf(fpout,"[%3.1f %3.1f] 0 d\n",rf[i].dash1,rf[i].dash2);
          else
            fprintf(fpout,"[] 0 d\n");
        }
        ps_line(fpout,startx,starty,laststartx,laststarty,tmp_connec_line);
        fprintf(fpout,"[] 0 d\n");  /* PS:setdash--always unset dash */
      }  /* if draw line */
    }  /* if LINE */

    /*** draw pnts (if pnts or lines+rfs, and RF, and !B.S.)--may be num'd */
    if ((plottype==PNTS || (plottype==RFS && lineflag)) &&
                           rf[i].radius_deg>0.0 && rf[i].x!=9999.0)  {
      if (plottype != PNTS && rf_whiteness > 0.0) /* no line variation */
        rf[i].colorflag = DARK;
      if (rf[i].colorflag==BLACK)  {
        tmp_pnt_dia = pnt_dia;
        tmp_pnt_line = 2.0*pnt_line;
        filled_flag = FILLED;
        tmp_pnt_fill = 0.0;
      }
      if (rf[i].colorflag==DARK)  {
        tmp_pnt_dia = pnt_dia - pnt_line;
        tmp_pnt_line = pnt_line;
        filled_flag = FILLED;
        tmp_pnt_fill = 1.0;
      }
      if (rf[i].colorflag==LIGHT)   {
        tmp_pnt_dia = pnt_dia;
        tmp_pnt_line = 0.5*pnt_line;
        filled_flag = UNFILLED;
        tmp_pnt_fill = 1.0;
      }
      if (rf[i].colorflag==NONE) { /* force values (overrides parmfile) */
        if (rf[i].pntfill>=0.0) filled_flag = FILLED;
        else                      filled_flag = UNFILLED;
        tmp_pnt_line = rf[i].pntline;
        tmp_pnt_fill = rf[i].pntfill;
        tmp_pnt_dia = pnt_dia;
      }
      ps_pnt(fpout,startx,starty,tmp_pnt_dia,tmp_pnt_line,filled_flag,
             tmp_pnt_fill);
      if (i && lineflag) {  /* if line + !first, redraw point to occlude */
        ps_pnt(fpout,laststartx,laststarty,lasttmp_pnt_dia,lasttmp_pnt_line,
               lastfilled_flag,lasttmp_pnt_fill); }
      lasttmp_pnt_dia = tmp_pnt_dia;  /* save for redraw next time */
      lasttmp_pnt_line = tmp_pnt_line;
      lastfilled_flag = filled_flag;
      lasttmp_pnt_fill = tmp_pnt_fill;
      if (numflag && plottype == PNTS)  {  /* draw point numbers */
        ps_text(fpout,rf[i].name,
                startx + 0.5*tmp_pnt_dia + 0.5*tmp_pnt_line,
                starty - 0.3*rf_fontsize,rf_fontsize);  }
    }  /* if PNT */

    /* save coords if RFthere (even if first and/or break) */
    if (rf[i].radius_deg>0.0)  { laststartx=startx; laststarty=starty; }

  } /* end of rfs/points/lines loop */

  /* draw hemifield on top of RFs */
  if (hemifield_front)
    ps_hemifield(fpout,center_x,center_y,max_rad,max_ecc,hf_line,
                 radians_to_pts,vf_fontsize,left_hemifield,quadrant_flag);

  /* Postscript trailer */
  fprintf(fpout,"showpage\n");
  fclose(fpout);
  exit(0);
}

/*** begin functions */
void ps_header(FILE *fpout)
{
  fprintf(fpout,"%%!PS-Adobe-2.0\n");
  fprintf(fpout,"%%%%BoundingBox: 0.0 0.0 612.0 792.0\n");
  fprintf(fpout,"1.00 1.00 scale\n");
  fprintf(fpout,"newpath\n");
  fprintf(fpout,"0.00 0.00 translate\n");

  /* defs for a smaller output file */
  fprintf(fpout,"/a { gsave } def\n");
  fprintf(fpout,"/c { scale } def\n");
  fprintf(fpout,"/d { setdash } def\n");
  fprintf(fpout,"/e { grestore } def\n");
  fprintf(fpout,"/f { fill } def\n");
  fprintf(fpout,"/g { setgray } def\n");
  fprintf(fpout,"/i { itransform } def\n");
  fprintf(fpout,"/l { lineto } def\n");
  fprintf(fpout,"/L { rlineto } def\n");
  fprintf(fpout,"/m { moveto } def\n");
  fprintf(fpout,"/M { rmoveto } def\n");
  fprintf(fpout,"/n { newpath } def\n");
  fprintf(fpout,"/p { closepath } def\n");
  fprintf(fpout,"/r { rotate } def\n");
  fprintf(fpout,"/s { stroke } def\n");
  fprintf(fpout,"/S { show } def\n");
  fprintf(fpout,"/t { translate } def\n");
  fprintf(fpout,"/w { setlinewidth } def\n");
  fprintf(fpout,"/x { transform } def\n");
  fprintf(fpout,"/md { /mtrx matrix def } def\n");
  fprintf(fpout,"/ms { /savematrix mtrx currentmatrix def } def\n");
  fprintf(fpout,"/mt { savematrix setmatrix } def\n");
  fprintf(fpout,"/fh { /Helvetica findfont } def\n");
  fprintf(fpout,"/sl { stringwidth pop 0 exch sub } def\n");
  fprintf(fpout,"/ss { scalefont setfont } def\n");
  fprintf(fpout,"/getyyxx { /yy exch def /xx exch def } def\n");

  /* bug in Illustrator leaves dash set */
  fprintf(fpout,"[] 0 d\n");  /* PS:setdash--unset dash */
}

void ps_parms(FILE *fpout, int argc, char **argv)
{
  int   i;
  char  line[200];
  FILE  *fpparm=NULL;

  fprintf(fpout,"%% cmdline:");
  for (i=0; i<argc; i++)
    fprintf(fpout," %s",argv[i]);
  fprintf(fpout,"\n");

  fpparm = fopen("arr.parm","r");
  if (fpparm==NULL) {
    fprintf(fpout,"%% no arr.parm: used all-default parms\n");
  }
  else {
    fprintf(fpout,"%% arr.parm (non-default parms):\n");
    while (fgets(line,80,fpparm) != NULL) {
      if (strlen(line) && line[0]!='\n') {
        fprintf(fpout,"%% %s",line);     /* fgets keeps newline */
        if (line[strlen(line)-1]!='\n')  /* if last missing */
          fprintf(fpout,"\n");
      }
    }
  }
}

void ps_hemifield(FILE *fpout, float center_x, float center_y, float max_rad,
                  float max_ecc, float hf_line, float radians_to_pts,
                  float vf_fontsize, int left_hemifield, int quadrant_flag)
{
  char   str[200];
  int    ecc;
  float  x, y;

  /* HM */
  if (left_hemifield) ps_rline(fpout,center_x,center_y,-max_rad,0.0,hf_line);
  else                ps_rline(fpout,center_x,center_y, max_rad,0.0,hf_line);

  /* VM */
  if (quadrant_flag==LOWER_QUAD)
    ps_rline(fpout,center_x,center_y-max_rad,0.0,max_rad,hf_line);
  if (quadrant_flag==BOTH_QUAD)
    ps_rline(fpout,center_x,center_y-max_rad,0.0,2.0*max_rad,hf_line);
  if (quadrant_flag==UPPER_QUAD)
    ps_rline(fpout,center_x,center_y,0.0,max_rad,hf_line);

  /* then isoeccentricity lines */
  ps_isoecc(fpout,center_x,center_y,max_ecc*1.0000*CONVERT*radians_to_pts,
            hf_line,left_hemifield,quadrant_flag);
  ps_isoecc(fpout,center_x,center_y,max_ecc*0.6666*CONVERT*radians_to_pts,
            hf_line,left_hemifield,quadrant_flag);
  ps_isoecc(fpout,center_x,center_y,max_ecc*0.3333*CONVERT*radians_to_pts,
            hf_line,left_hemifield,quadrant_flag);

  /* label them */
  if (left_hemifield) x = center_x + 0.5*vf_fontsize;
  else                x = center_x - 2.0*vf_fontsize;
  y = center_y - 0.5*vf_fontsize;
  if (quadrant_flag==UPPER_QUAD || quadrant_flag==BOTH_QUAD) {
    ecc = (int)max_ecc;
    sprintf(str,"+%d",ecc);
    ps_text(fpout,str,x,y+(float)ecc*CONVERT*radians_to_pts,vf_fontsize);
    ecc = (int)max_ecc*2/3;
    sprintf(str,"+%d",ecc);
    ps_text(fpout,str,x,y+(float)ecc*CONVERT*radians_to_pts,vf_fontsize);
    ecc = (int)max_ecc/3;
    sprintf(str,"+%d",ecc);
    ps_text(fpout,str,x,y+(float)ecc*CONVERT*radians_to_pts,vf_fontsize);
  }
  if (quadrant_flag==LOWER_QUAD || quadrant_flag==BOTH_QUAD) {
    ecc = (int)max_ecc/3;
    sprintf(str,"-%d",ecc);
    ps_text(fpout,str,x,y-(float)ecc*CONVERT*radians_to_pts,vf_fontsize);
    ecc = (int)max_ecc*2/3;
    sprintf(str,"-%d",ecc);
    ps_text(fpout,str,x,y-(float)ecc*CONVERT*radians_to_pts,vf_fontsize);
    ecc = (int)max_ecc;
    sprintf(str,"-%d",ecc);
    ps_text(fpout,str,x,y-(float)ecc*CONVERT*radians_to_pts,vf_fontsize);
  }
}

/*============================================================================
 Write postscript to draw RF ellipse at (center_x,center_y).  Needs (r,theta)
 for flat correction.  Whiteness is 0.0-1.0 (B->W),neg is no fill (clear).
============================================================================*/
void ps_rf(FILE *fpout, float cent_x, float cent_y, float r, float theta,
           float halflen, float halfwi, float angle, float thick,
           float whiteness, int flatflg)
{
  fprintf(fpout,"n\n");        /* PS:newpath */
  fprintf(fpout,"md\n");       /* PS:matrix */
  fprintf(fpout,"ms\n");       /* PS:currentmatrix (save matrix) */

  /* translate to RF center */
  fprintf(fpout,"%3.2f %3.2f t\n",cent_x,cent_y);  /* PS:translate */

  /* correct for flat hemisphere first */
  if (flatflg) {
    fprintf(fpout,"%3.2f r\n",theta);  /* PS:rotate */
    fprintf(fpout,"%3.2f %3.2f c\n",1.0,1.0+(r/90.0*(M_PI_2-1.0))); /*scale*/
    /* undo theta rotation so following ellipse scale is in right direction */
    fprintf(fpout,"%3.2f r\n",-theta);  /* PS:rotate */
  }

  /* then scale for ellipse */
  fprintf(fpout,"%3.2f r\n",angle);                 /* PS:rotate */
  fprintf(fpout,"%3.2f %3.2f c\n",halflen,halfwi);  /* PS:scale */
  fprintf(fpout,"%1.2f w\n",thick);                 /* PS:setlinewidth */

  /* draw RF circle */
  fprintf(fpout,"0.0 0.0 1.0 0.0 360.0 arc\n");
  fprintf(fpout,"mt\n");       /* PS:setmatrix (restore former matrix) */

  /* fill RF first if needed */
  if (whiteness>=0.0)  {
    fprintf(fpout,"%3.2f g\n",whiteness);  /* PS:setgray */
    fprintf(fpout,"a\n");      /* PS:gsave (save rf boundary path) */
    fprintf(fpout,"f\n");      /* PS:fill */
    fprintf(fpout,"e\n");      /* PS:grestore */
    fprintf(fpout,"0.0 g\n");  /* PS:setgray */
  }

  /* stroke circle afterward so RF border is visible */
  fprintf(fpout,"s\n");        /* PS:stroke */
}

/*============================================================================
 Write Postscript to draw RF label with line to edge of ellipse (x,y=RFcent)
============================================================================*/
void ps_rflabel(FILE *fpout, float x, float y, float r, float theta,
                float h_len, float h_wi, float angle, float thick,
                char *str, float font, int flatflg, int left_hemi,
                float label_angle_deg, float stalk_ratio)
{
  float  label_angle;

  /* label text adjustments fail outside -60 to 60 deg */
  if (label_angle_deg> 60.0)  label_angle_deg = 60.0;
  if (label_angle_deg< -60.0) label_angle_deg = -60.0;

  fprintf(fpout,"n\n");
  fprintf(fpout,"a\n");

  /* translate to RF center */
  fprintf(fpout,"%3.2f %3.2f t\n",x,y);

  /* correct for flat hemisphere first */
  if (flatflg) {  /* stretch by PI/2 at 90 deg */
    fprintf(fpout,"%3.2f r\n",theta);
    fprintf(fpout,"%3.2f %3.2f c\n",1.0,1.0+(r/90.0*(M_PI_2-1.0)));
    /* undo theta rotation so following ellipse scale is in right direction */
    fprintf(fpout,"%3.2f r\n",-theta);
  }

  /* do ellipse scaling */
  fprintf(fpout,"%3.2f r\n",angle);
  fprintf(fpout,"%3.2f %3.2f c\n",h_len,h_wi);

  /* undo ellipse rotation to get location for label line */
  if (left_hemi) fprintf(fpout,"%3.2f r\n",-angle + 180.0-label_angle_deg);
  else           fprintf(fpout,"%3.2f r\n",-angle + label_angle_deg);

  /* put origin where we want the line/text (RF is transformed radius=1 arc) */
  fprintf(fpout,"1.0 0.0 t\n");

  /* get page coord where we are (so we can pop to avoid scaling text,lines) */
  fprintf(fpout,"0 0 x\n");   /* save where we are */
  fprintf(fpout,"e\n");       /* pop to untransformed */
  fprintf(fpout,"i\n");       /* get the diff */
  fprintf(fpout,"getyyxx\n");   /* save as yy xx */
  fprintf(fpout,"xx yy m\n"); /* move there in untransformed */

  /* draw line */
  fprintf(fpout,"0.0 g\n");
  fprintf(fpout,"%1.2f w\n",thick);
  label_angle = label_angle_deg*CONVERT;
  if (left_hemi)
    fprintf(fpout,"%3.2f %3.2f L\n",-stalk_ratio*font*cos(label_angle),
                          stalk_ratio*font*TOWARD_HOR*sin(label_angle));
  else
    fprintf(fpout,"%3.2f %3.2f L\n",stalk_ratio*font*cos(label_angle),
                            stalk_ratio*font*TOWARD_HOR*sin(label_angle));

  /* write text (fine adjustments) */
  fprintf(fpout,"fh\n");
  fprintf(fpout,"%d ss\n",(int)font);
  if (left_hemi) {
    fprintf(fpout,"(%s) sl\n",str);          /* x (-stringlength) on stack */
    fprintf(fpout,"%3.2f sub\n",font/15.0);  /* a little extra -x space */
    fprintf(fpout,"%3.2f M\n",
          -((45.0-label_angle_deg)/65.0) * (font/2.0));  /* y-adjust hell */
  }
  else {
    fprintf(fpout,"%3.2f %3.2f M\n",font/15.0,
          -((45.0-label_angle_deg)/65.0) * (font/2.0));  /* ditto */
  }
  fprintf(fpout,"(%s) S\n",str);
  fprintf(fpout,"s\n"); /* last because stroke does newpath */
}

void ps_text(FILE *fpout, char *str, float x, float y, float fontsize)
{
  fprintf(fpout,"0.0 g\n");
  fprintf(fpout,"fh\n");
  fprintf(fpout,"%d ss\n",(int)fontsize);
  fprintf(fpout,"%3.2f %3.2f m\n",x,y);
  fprintf(fpout,"(%s) S\n",str);
}

void ps_pnt(FILE *fpout, float center_x, float center_y, float dia,
            float thickness, int filled_flag, float filled_color)
{
  fprintf(fpout,"n\n");          /* PS: newpath */
  fprintf(fpout,"0.0 g\n");      /* PS: setgray */
  fprintf(fpout,"%1.2f w\n",thickness);  /* PS: setlinewitdh */
  fprintf(fpout,"%3.2f %3.2f %3.2f 0.0 360.0 arc\n",
                    center_x,center_y,dia/2.0);
  fprintf(fpout,"p\n");          /* PS: closepath */
  if (filled_flag) {
    fprintf(fpout,"%3.2f g\n",filled_color);  /* PS: setgray */
    fprintf(fpout,"a\n");        /* PS: gsave (save path for stroke) */
    fprintf(fpout,"f\n");        /* PS: fill */
    fprintf(fpout,"e\n");        /* PS: grestore */
    fprintf(fpout,"0.0 g\n");    /* PS: setgray (undo for next) */
  }
  fprintf(fpout,"s\n");
}

void ps_rline(
  FILE *fpout, float x, float y, float dx, float dy, float thickness)
{
  fprintf(fpout,"n\n");
  fprintf(fpout,"%1.2f w\n",thickness);
  fprintf(fpout,"%3.2f %3.2f m\n",x,y);
  fprintf(fpout,"%3.2f %3.2f L\n",dx,dy);
  fprintf(fpout,"s\n");
}

void ps_line(
  FILE *fpout, float x1, float y1, float x2, float y2, float thickness)
{
  fprintf(fpout,"n\n");
  fprintf(fpout,"%1.2f w\n",thickness);
  fprintf(fpout,"%3.2f %3.2f m\n",x1,y1);
  fprintf(fpout,"%3.2f %3.2f l\n",x2,y2);
  fprintf(fpout,"s\n");
}

void ps_isoecc(FILE *fpout, float centx, float centy,float radius,
               float thickness, int left_hemifield_flag, int quadrant_flag)
{
  fprintf(fpout,"n\n");
  fprintf(fpout,"0.0 g\n");
  fprintf(fpout,"%1.2f w\n",thickness);
  if (left_hemifield_flag) {
    if (quadrant_flag==LOWER_QUAD)
      fprintf(fpout,"%3.2f %3.2f %3.2f 180.0 270.0 arc\n",centx,centy,radius);
    if (quadrant_flag==BOTH_QUAD)
      fprintf(fpout,"%3.2f %3.2f %3.2f 90.0 270.0 arc\n",centx,centy,radius);
    if (quadrant_flag==UPPER_QUAD)
      fprintf(fpout,"%3.2f %3.2f %3.2f 90.0 180.0 arc\n",centx,centy,radius);
  }
  else {
    if (quadrant_flag==LOWER_QUAD)
      fprintf(fpout,"%3.2f %3.2f %3.2f 270.0 0.0 arc\n",centx,centy,radius);
    if (quadrant_flag==BOTH_QUAD)
      fprintf(fpout,"%3.2f %3.2f %3.2f 270.0 90.0 arc\n",centx,centy,radius);
    if (quadrant_flag==UPPER_QUAD)
      fprintf(fpout,"%3.2f %3.2f %3.2f 0.0 90.0 arc\n",centx,centy,radius);
  }
  fprintf(fpout,"s\n");
}

void rotate_x(float r[3], float angle)  /* rotate vect around x-axis radians */
{
  float  y, z;

  y = r[1];
  z = r[2];
  r[1] = y*cos((double)angle) + z*sin((double)angle);
  r[2] = -y*sin((double)angle) + z*cos((double)angle);
}

void rotate_y(float r[3],float angle)  /* rotate vect around y-axis radians */
{
  float  x, z;

  x = r[0];
  z = r[2];
  r[0] = x*cos((double)angle) - z*sin((double)angle);
  r[2] = x*sin((double)angle) + z*cos((double)angle);
}

void rotate_z(float r[3], float angle)  /* rotate vect around z-axis radians */
{
  float  x, y;

  x = r[0];
  y = r[1];
  r[0] = x*cos((double)angle) + y*sin((double)angle);
  r[1] = -x*sin((double)angle) + y*cos((double)angle);
}

void spherpolar_to_cartesian(float r[3])  /* spherpol (rho,th,phi) -> x,y,z */
{
  float  rho, theta, phi;

  rho = r[0];
  theta = r[1];
  phi = r[2];
  r[0] = (double)rho * sin((double)phi) * cos((double)theta);
  r[1] = (double)rho * sin((double)phi) * sin((double)theta);
  r[2] = (double)rho * cos((double)phi);
}

void cartesian_to_spherpolar(float r[3])  /* x,y,z -> spherpol (rho,th,phi) */
{
  float  x, y, z, rho, theta=0.0, phi;

  x = r[0];
  y = r[1];
  z = r[2];
  rho = sqrt((double)(x*x + y*y + z*z));
  if (rho==0.0)
    theta = phi = 0.0;
  else {
    if (x>0.0)
      theta = atan((double)(y/x));
    if (x<0.0)
      theta = M_PI + atan((double)(y/x));
    if (x==0.0 && y>=0.0)
      theta = M_PI_2;
    if (x==0.0 && y<0.0)
      theta = 3.0 * M_PI_2;
    phi = acos((double)(z/(rho)));
  }
  r[0] = rho;
  r[1] = theta;
  r[2] = phi;
}

void print_help(void)
{
printf("\
#############################################################################\n\
              rfplots -- make PostScript receptive fields plot\n\
\n\
                       written by Marty Sereno\n\
                  University of California, San Diego\n\
#############################################################################\n\
    Reads ASCII data file of penetration coordinates and RF params\n\
  and writes Postscript file with hemifield and receptive fields.  The\n\
  first flag chooses receptive field ellipses vs. points; the second flag\n\
  determines whether lines should be drawn between successive receptive\n\
  field centers (for rfs plus lines, receptive field centers are also draw);\n\
  the third flag determines whether the receptive fields or points should\n\
  be numbered.\n\
\n\
    Format for data line in input file (which is also read by arrows.c)\n\
  is either 6 or 8 args:\n\
\n\
pointname  point-x  point-y  RF-r  RF-theta  RF-dia       \n\
                        *or*\n\
pointname  point-x  point-y  RF-r  RF-theta  RF-length  RF-width  RF-angle\n\
\n\
    RF-r (distance from center of gaze to receptive field center)\n\
  given in 0-180+ degrees and RF-theta (angle of center-of-gaze-to-RF\n\
  line) in -180 to +180 degrees, RF-dia (or RF-length, RF-width)\n\
  given in positive degrees and RF-angle (measured between the long \n\
  axis of the receptive field and horizontal) given in -90 to +90 degrees.\n\
  There can be up to 10 chars in point name.\n\
\n\
    Right and left hemifield data should be entered with the horizontal\n\
  meridian as zero degrees.  Default is to draw arrows or receptive\n\
  fields as a right hemifield.  If left hemifield is desired, set\n\
  LEFT_HEMIFIELD flag.\n\
\n\
    Lines with leading pound sign or space are treated as comments\n\
  and skipped (i.e., data or command lines *must* have a character \n\
  in the first column).  Can't deal with comments beginning midline.\n\
\n\
    Commands have leading (first column) slash.  These are all toggles\n\
  that apply to all RFs that follow until next command is read.  Commands\n\
  currently understood are:\n\
\n\
/break\n\
/black\n\
/dark\n\
/light\n\
/rfline <arg>\n\
/connecline <arg>\n\
/pntline <arg>\n\
/labelline <arg>\n\
/dash <arg> <arg>\n\
/rffill <arg>\n\
/pntfill <arg>\n\
\n\
    Command /break prevents line between RFs in line mode (ignored in rfs\n\
  mode).  Commands /black, /dark, and /light adjust RF, point, connec line\n\
  widths, and RF fills (depending on context and parmfile values).  Values\n\
  can be set explicitly with multiple-argument commands: /rfline <arg>,\n\
  /connecline <arg>, /pntline <arg>, /labelline <arg>, /dash <arg> <arg>,\n\
  /rffill <arg>, and /pntfill <arg>.  Line and dash command args are in\n\
  points.  To unset dash, give negative args.  Fill arguments are 0.0 => 1.0\n\
  for black => opaque white and -1.0 for clear.  NOTE: setting explicit\n\
  values with these commands in datafile will override (ignore) settings in\n\
  parmfile.  If one or more values are explicitly set, other unset ones will\n\
  default to values in parmfile.\n\
\n\
#\n\
# Example data file\n\
# \n\
/rfline 0.5\n\
/rffill 0.8\n\
12F	5.23	9.25	24	-25	7.5	4.5	50\n\
254-L	2.31	5.36	37	50	5.5	3.5	-60\n\
312G	10.55	4.46	10	-10	7.5	5	-40\n\
#\n\
# no-response point (no RF; point but no arrow)\n\
25NR	6.24	10.21	-1	0	0	0	0\n\
#\n\
#\n\
/break\n\
/rfline 0.8\n\
/dash 3 2\n\
571	3.24	7.35	12	57	14	9	-27\n\
601B	6.01	10.05	17	-9	10	8	-5\n\
#\n\
/dash -1 -1\n\
# blind spot 'receptive field' (RF; no point or arrow)\n\
/break\n\
/dark\n\
BS-1	9999.0	9999.0	21	-5	4	2	90\n\
#\n\
#\n\
    Also reads parm file (rfs.parm) file containing various settings\n\
  (e.g., margins, postscript lines sizes, coordinate transforms,\n\
  flat correct flag); if not there writes a default version of\n\
  the file, which can then be edited.  Ignores properly formatted \n\
  parameters in parm file that it doesn't recognize.  If a parm appears\n\
  more than once, the last instance determines its value.  Parms omitted\n\
  from file are set to default values.  As noted above, explicit commands \n\
  to set rf, connec, and point line, and rf and pnt fill in data file\n\
  will override settings in parmfile.  Cmdline and non-default parms are\n\
  written as comments at head of PostScript outfile. Default parm file:\n\
\n\
  Default parm file:\n\
\n\
RF_LINE = 0.40\n\
PNT_LINE = 0.40\n\
CONNEC_LINE = 0.40\n\
PNT_DIA = 4.00\n\
HF_LINE = 0.20\n\
RF_FONTSIZE = 9.00\n\
RF_LABEL_ANGLE = 30.00\n\
RAND_WIDTH = 0.00\n\
STALK_RATIO = 0.8\n\
VF_FONTSIZE = 12.00\n\
MAX_ECC = 45.00\n\
FLAT_CORRECT = 1\n\
MARG = 30.00\n\
MOVE_ON_HM = 0.00\n\
MOVE_ON_VM = 0.00\n\
ROTATE_ON_CG = 0.00\n\
LEFT_HEMIFIELD = 0\n\
RF_WHITENESS = -1.0\n\
RANDOM_COLOR = 0\n\
HEMIFIELD_FRONT = 0\n\
LEGEND_FLAG = 1\n\
QUADRANT_FLAG = 0\n\
\n\
  Calculates receptive fields (ignores x-y location of recording site),\n\
  generates PS-Adobe-2.0 format Postscript file to print rfs in\n\
  regular/upright format.  Accepts up to 5000 rfs.  Writes out the\n\
  penetration number in RF_FONTSIZE pt Helvetica if requested at edge of\n\
  RF, or next to point.  Draws 3 isoeccentricity lines (set max\n\
  isoeccentricity with MAX_ECC).  Corrects for effect of flattening\n\
  hemifield (since radial measurements are unchanged, there is a PI/2\n\
  circumferential stretching of RF centered at 90 degrees)--use\n\
  FLAT_CORRECT flag.\n\
\n\
    Put a negative number for r (and fill remaining entries with 0's)\n\
  to get a point with no RF (e.g., injection, lesion, no response).\n\
  Put in 9999.0 for x and y to suppress arrow but still get a pseudo-RF\n\
  (e.g., for blind spot measurments).  These 'RFs' are never labeled.\n\
\n\
    Can rotate coord system of RFs.  This is done in the following order:\n\
  move along the HM (rotate around cartesian y-axis), move along the\n\
  VM (rotate around cartesian x-axis), and rotate around the center of\n\
  gaze (rotate around cartesian z-axis).\n\
\n\
    Set receptive field fill with RF_WHITENESS: neg num is clear while\n\
  0.0 to 1.0 is black to opaque white.  RF_WHITENESS sets color of /dark\n\
  (default) colored RFs; /light and /black RFs are lighter (dashed line)\n\
  and darker (thicker line).  If receptive fields are colored, line\n\
  variation (dashed, thickness) is overriden (all are set to /dark).  \n\
  Setting RANDOM_COLOR makes randomly colored RFs.  Setting RF color with\n\
  explicit /commands will override parmfile settings.\n\
\n\
    RF_LABEL_ANGLE sets angle of label line (esp. to avoid occlusion with\n\
  opaque RFs).  Zero degrees is horizontal with label at periphery.  Angle\n\
  restricted to -60 to 60 degrees--i.e., labels are always peripheral.\n\
  STALK_RATIO is length of stalk relative to fontsize.  RAND_WIDTH sets\n\
  the width of random jitter of label angle (for dense RF plots).\n\
\n\
    HEMIFIELD_FRONT puts hemifield in front (for when there are many opaque\n\
  RFs).  Unset LEGEND_FLAG to get rid of legend.  QUADRANT_FLAG (-1/0/1)\n\
  sets upper (1), lower (-1), or both (0) without re-centering.\n\
\n\
    Ignores 3 arg parms in datafile (used by arrows.c)\n\
#############################################################################\n\
\n\
");
}

